"""Tools for text."""

from enum import IntFlag
from typing import Final

import numpy as np
from numpy.typing import NDArray
from ugrapheme import grapheme_iter
from uwcwidth import wcswidth

from ._batgrl_markdown import find_md_tokens
from .colors import BLACK, WHITE, Color
from .geometry import Size
from .logging import get_logger

__all__ = [
    "Cell",
    "Style",
    "add_text",
    "is_word_char",
    "new_cell",
    "smooth_horizontal_bar",
    "smooth_vertical_bar",
]

EGC_BASE: Final = 0x180000
"""A bit flag to mark egcs in canvas arrays. Must be greater than `sys.maxunicode`."""
EGC_POOL: list[str] = []
"""
Storage for extended grapheme clusters.

If `ord_` is an ord in a canvas array and `ord_ & EGC_BASE == 1`, then `ord_ - EGC_BASE`
is an index into `EGC_POOL`. In this way, we can use the uint32 "ord" field in canvas
arrays to store both codepoints and egcs.
"""
EGCS: dict[str, int] = {}
"""Extended grapheme clusters currently stored in EGC_POOL and their index."""
VERTICAL_BLOCKS: Final = " ▁▂▃▄▅▆▇█"
HORIZONTAL_BLOCKS: Final = " ▏▎▍▌▋▊▉█"

logger = get_logger(__name__)


class Style(IntFlag):
    """The graphic rendition parameters of a terminal cell."""

    # ! Don't use auto() as these must match the definitions in _rendering.pyx.
    DEFAULT = 0
    BOLD = 0b1
    ITALIC = 0b10
    UNDERLINE = 0b100
    STRIKETHROUGH = 0b1000
    OVERLINE = 0b10000
    REVERSE = 0b100000


Cell = np.dtype(
    [
        ("ord", "ulong"),
        ("style", "u1"),
        ("fg_color", "u1", (3,)),
        ("bg_color", "u1", (3,)),
    ]
)
"""A structured array type that represents a single cell in a terminal."""


def egc_ord(text: str) -> int:
    """
    Return either a unicode codepoint or an index into the egc pool for the first
    extended grapheme cluster in ``text``.

    Parameters
    ----------
    text : str
        An extended grapheme cluster.

    Returns
    -------
    int
        A unicode codepoint or an index into ``EGC_POOL``.

    See Also
    --------
    egc_chr
    """
    if len(text) == 0:
        return 0

    egc = next(grapheme_iter(text))
    if len(egc) == 1:
        return ord(egc)

    if egc not in EGCS:
        # FIXME: Unbounded growth
        logger.debug(f"EGC Added: {egc} {[ord(i) for i in egc]}")
        EGCS[egc] = len(EGC_POOL)
        EGC_POOL.append(egc)
    return EGCS[egc] | EGC_BASE


def egc_chr(ord: int) -> str:
    """
    Return the extended grapheme cluster represented by ``ord``.

    Parameters
    ----------
    ord : int
        The ord of the extended grapheme cluster.

    Returns
    -------
    str
        The extended grapheme cluster represented by ``ord``.

    See Also
    --------
    egc_ord
    """
    return EGC_POOL[ord - EGC_BASE] if ord & EGC_BASE else chr(ord)


def is_word_char(char: str) -> bool:
    """
    Whether `char` is a word character.

    A character is a word character if it is alphanumeric or an underscore.

    Parameters
    ----------
    char : str
        The char to test.

    Returns
    -------
    bool
        Whether the char is a word character.
    """
    return char.isalnum() or char == "_"


def new_cell(
    ord: int = 0x20,
    style: Style = Style.DEFAULT,
    fg_color: Color = WHITE,
    bg_color: Color = BLACK,
) -> NDArray[Cell]:
    """
    Create a ``Cell`` scalar.

    A Cell is a structured array type that represents a single cell in a terminal.

    Parameters
    ----------
    ord : int, default: 0x20
        The cell's character's ord.
    style : Style, Style.DEFAULT
        The style (bold, italic, etc.) of the cell.
    fg_color : Color, default: WHITE
        Foreground color of cell.
    bg_color : Color, default: BLACK
        Background color of cell.

    Returns
    -------
    NDArray[Cell]
        A ``Cell`` scalar.
    """
    return np.array((ord, style, fg_color, bg_color), dtype=Cell)


def _parse_batgrl_md(text: str) -> tuple[Size, list[list[NDArray[Cell]]]]:
    """
    Parse batgrl markdown and return the minimum canvas size to fit text and
    a list of lines of styled characters.

    #### Syntax for batgrl markdown
    - italic: `*this is italic text*`
    - bold: `**this is bold text**`
    - strikethrough: `~~this is strikethrough text~~`
    - underlined: `__this is underlined text__`
    - overlined: `^^this is overlined text^^`

    Parameters
    ----------
    text : str
        The text to parse.

    Returns
    -------
    tuple[Size, list[list[Cell]]]
        Minimum canvas size to fit text and a list of lines of styled characters.
    """
    NO_CHAR = new_cell(ord=0)
    matches, escapes = find_md_tokens(text)
    cells = [
        new_cell(ord=egc_ord(egc))[["ord", "style"]] for egc in grapheme_iter(text)
    ]
    for before, start, end, after, style in matches:
        cells[start - before : start] = [NO_CHAR] * before
        cells[end : end + after] = [NO_CHAR] * after
        for i in range(start, end):
            cells[i]["style"] |= getattr(Style, style.upper())

    for i in escapes:
        cells[i] = NO_CHAR

    lines = []
    line = []
    line_width = max_width = 0

    for cell in cells:
        ord_ = cell["ord"].item()
        if ord_ == 0:
            continue

        if ord_ == 0xA:
            lines.append(line)
            line = []
            if line_width > max_width:
                max_width = line_width
            line_width = 0
        else:
            width = wcswidth(egc_chr(ord_))
            if width > 0:
                line_width += width
                line.append(cell)

    lines.append(line)
    if line_width > max_width:
        max_width = line_width

    return Size(len(lines), max_width), lines


def _text_to_cells(text: str) -> tuple[Size, list[list[NDArray[Cell]]]]:
    """
    Convert some text to a list of lists of Cells and the minimum canvas size to fit
    them.

    Parameters
    ----------
    text : str
        The text to convert.

    Returns
    -------
    tuple[Size, list[list[Cell]]]
        Minimum canvas size to fit text and a list of lists of Cells.
    """
    egcs = [list(grapheme_iter(line)) for line in text.split("\n")]
    cells = []
    max_width = 0
    for egc_line in egcs:
        line_width = 0
        cell_line = []
        for egc in egc_line:
            egc_width = wcswidth(egc)
            if egc_width <= 0:
                continue
            cell_line.append(new_cell(ord=egc_ord(egc)))
            line_width += egc_width
        cells.append(cell_line)
        if line_width > max_width:
            max_width = line_width
    return Size(len(cells), max_width), cells


def _write_cells_to_canvas(cells, canvas, fg_color, bg_color) -> None:
    """Write a list of lists of Cells to a canvas array."""
    _, columns = canvas.shape
    for cell_line, canvas_line in zip(cells, canvas):
        i = 0
        for cell in cell_line:
            char_width = wcswidth(egc_chr(cell["ord"].item()))
            if char_width < 1:
                continue

            if i + char_width > columns:
                canvas_line[i:]["ord"] = 0x2
                canvas_line[i:]["style"] = 0
                if fg_color is not None:
                    canvas_line[i:]["fg_color"] = fg_color
                if bg_color is not None:
                    canvas_line[i:]["bg_color"] = bg_color
                break

            canvas_line[i]["ord"] = cell["ord"]
            canvas_line[i + 1 : i + char_width]["ord"] = 0
            canvas_line[i : i + char_width]["style"] = cell["style"]

            if fg_color is not None:
                canvas_line[i : i + char_width]["fg_color"] = fg_color
            if bg_color is not None:
                canvas_line[i : i + char_width]["bg_color"] = bg_color

            i += char_width
            if i >= columns:
                break


def put_egc(canvas: NDArray[Cell], text: str) -> None:
    """
    Set each ord in canvas to represent the first extended grapheme cluster in ``text``.

    Parameters
    ----------
    canvas : NDArray[Cell]
        A ``Cell`` array or view.
    text : str
        An extended grapheme cluster.
    """
    canvas[:] = egc_ord(text)


def add_text(
    canvas: NDArray[Cell],
    text: str,
    *,
    fg_color: Color | None = None,
    bg_color: Color | None = None,
    markdown: bool = False,
    truncate_text: bool = False,
) -> None:
    """
    Add multiple lines of text to a view of a canvas.

    If `markdown` is true, text can be styled using batgrl markdown. The syntax is:
    - italic: `*this is italic text*`
    - bold: `**this is bold text**`
    - strikethrough: `~~this is strikethrough text~~`
    - underlined: `__this is underlined text__`
    - overlined: `^^this is overlined text^^`

    Text is added starting at first index in canvas. Every new line is added on a new
    row.

    Parameters
    ----------
    canvas : NDArray[Cell]
        A 1- or 2-dimensional view of a ``Cell`` array.
    text : str
        Text to add to canvas.
    fg_color : Color | None, default: None
        Foreground color of text. If not given, current canvas foreground color is used.
    bg_color : Color | None, default: None
        Background color of text. If not given, current canvas background color is used.
    markdown : bool, default: False
        Whether to parse text for batgrl markdown.
    truncate_text : bool, default: False
        For text that doesn't fit on canvas, truncate text if true else raise an
        `IndexError`.
    """
    size, cells = _parse_batgrl_md(text) if markdown else _text_to_cells(text)
    if canvas.ndim == 1:  # Pre-pend an axis if canvas is one-dimensional.
        canvas = canvas[None]
    rows, columns = canvas.shape
    if not truncate_text and (size.height > rows or size.width > columns):
        raise IndexError("Text does not fit in canvas.")
    _write_cells_to_canvas(cells, canvas, fg_color, bg_color)


def canvas_as_text(canvas: NDArray[Cell], line_widths: list[int] | None = None) -> str:
    """
    Return a ``Cell`` array as a single multi-line string.

    Parameters
    ----------
    canvas : NDArray[Cell]
        The ``Cell`` array to convert.
    line_widths : list[int] | None
        Optionally truncate line ``n`` to have column width ``line_widths[n]``. If
        line_widths[``n``] is greater than the column width of line ``n`` it is ignored.

    Returns
    -------
    str
        The canvas as a multi-line string.
    """
    if canvas.ndim == 1:  # Pre-pend an axis if canvas is one-dimensional.
        canvas = canvas[None]
    rows, columns = canvas.shape
    if line_widths is None:
        line_widths = [columns] * rows
    elif len(line_widths) < rows:
        line_widths = line_widths + [columns] * (rows - len(line_widths))

    text = []
    for row, line_width in zip(canvas, line_widths, strict=True):
        current_line_width = 0
        line = []
        for cell in row:
            char = egc_chr(cell["ord"])
            char_width = wcswidth(char)
            if char_width <= 0:
                continue
            if char_width + current_line_width > line_width:
                break
            current_line_width += char_width
            line.append(char)
        text.append("".join(line))
    return "\n".join(text)


def _smooth_bar(
    blocks: str,
    max_length: int,
    proportion: float,
    offset: float,
) -> tuple[str, ...]:
    """Create a smooth bar with given blocks."""
    if offset >= 1 or offset < 0:
        raise ValueError(
            f"Offset should greater than or equal to 0 and less than 1, but {offset} "
            "was given."
        )

    block_indices = len(blocks) - 1
    fill, partial = divmod(proportion * max_length, 1)
    fill = int(fill)

    if offset == 0.0:
        if fill == max_length:
            return (blocks[-1],) * fill

        index_partial = round(partial * block_indices)
        partial_block = blocks[index_partial]
        return (*(blocks[-1],) * fill, partial_block)

    partial += offset
    if partial > 1:
        partial -= 1
    else:
        fill -= 1

    index_offset = round(offset * block_indices)
    index_partial = round(partial * block_indices)
    offset_block = blocks[index_offset]
    partial_block = blocks[index_partial]
    return (offset_block, *(blocks[-1],) * fill, partial_block)


def smooth_vertical_bar(
    max_height: int,
    proportion: float,
    offset: float = 0.0,
    reversed: bool = False,
) -> tuple[str, ...]:
    """
    Create a vertical bar that's some proportion of max_height at an offset.

    Offset bars will return a minimum of 2 characters and the first character of the bar
    should have it's colors reversed (or, if bar is reversed, all colors should be
    reversed except first character).

    Parameters
    ----------
    max_height : int
        The height of the bar if proportion was 1.0.
    proportion : float
        The height of the bar as a proportion of the max_height.
    offset : float, default: 0.0
        Offset the bar vertically by some non-negative amount.
    reversed : bool, default: False
        Reversed vertical bar is drawn top-to-bottom and offset downwards.

    Returns
    -------
    tuple[str, ...]
        The bar as a tuple of characaters.
    """
    blocks = VERTICAL_BLOCKS[::-1] if reversed else VERTICAL_BLOCKS
    return _smooth_bar(blocks, max_height, proportion, offset)


def smooth_horizontal_bar(
    max_width: int, proportion: float, offset: float = 0.0
) -> tuple[str, ...]:
    """
    Create a horizontal bar that's some proportion of max_width at an offset.

    Offset bars will return a minimum of 2 characters and the first character of the bar
    should have it's colors reversed.

    Parameters
    ----------
    max_width : int
        The width of the bar if the proportion was 1.0.
    proportion : float
        The width of the bar as a proportion of the max_width.
    offset : float, default: 0.0
        Offset the bar horizontally by some non-negative amont.

    Returns
    -------
    tuple[str, ...]
        The bar as a tuple of characters.
    """
    return _smooth_bar(HORIZONTAL_BLOCKS, max_width, proportion, offset)



"""A graphic gadget."""

from pathlib import Path
from typing import Final, Literal

import cv2
import numpy as np
from numpy.typing import NDArray

from .._rendering import graphics_render
from ..colors import TRANSPARENT, AColor
from ..logging import get_logger
from ..texture_tools import Interpolation, resize_texture
from .gadget import Cell, Gadget, Point, PosHint, Size, SizeHint, bindable, clamp

__all__ = ["Blitter", "Graphics", "Interpolation", "Point", "Size", "scale_geometry"]

Blitter = Literal["braille", "full", "half", "octant", "quadrant", "sextant", "sixel"]
"""Determines how graphics are rendered."""

_BLITTER_GEOMETRY: Final[dict[Blitter, Size]] = {
    "braille": Size(4, 2),
    "full": Size(1, 1),
    "half": Size(2, 1),
    "octant": Size(4, 2),
    "quadrant": Size(2, 2),
    "sextant": Size(3, 2),
    "sixel": Size(20, 10),
}

logger = get_logger(__name__)


def scale_geometry[T: (Point, Size)](blitter: Blitter, point_or_size: T) -> T:
    """
    Scale a point or size by some blitter geometry.

    Parameters
    ----------
    blitter : Blitter
        Blitter from which pixel geometry is chosen.
    point_or_size : T
        A point or size to scale.

    Returns
    -------
    T
        The scaled geometry.
    """
    h, w = _BLITTER_GEOMETRY[blitter]
    a, b = point_or_size
    if blitter == "sixel":
        ah, _ = Graphics._sixel_aspect_ratio
        return type(point_or_size)(h * a // ah, w * b)
    return type(point_or_size)(h * a, w * b)


class Graphics(Gadget):
    r"""
    A graphic gadget. Displays arbitrary RGBA textures.

    Graphic gadgets' color information is stored in a uint8 RGBA array, :attr:`texture`.
    The size of :attr:`texture` depends on the geometry of the chosen blitter. For
    instance, if the chosen blitter is "half", then the texture will be twice the height
    of the gadget and the same width.

    Parameters
    ----------
    default_color : AColor, default: AColor(0, 0, 0, 0)
        Default texture color.
    alpha : float, default: 1.0
        Transparency of gadget.
    interpolation : Interpolation, default: "linear"
        Interpolation used when gadget is resized.
    blitter : Blitter, default: "half"
        Determines how graphics are rendered.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: True
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    texture : NDArray[np.uint8]
        uint8 RGBA color array.
    default_color : AColor
        Default texture color.
    alpha : float
        Transparency of gadget.
    interpolation : Interpolation
        Interpolation used when gadget is resized.
    blitter : Blitter
        Determines how graphics are rendered.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    to_png(path)
        Write :attr:`texture` to provided path as a `png` image.
    clear()
        Fill texture with default color.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    _sixel_support: bool = False
    """Whether sixel is supported."""
    _sixel_aspect_ratio: Size = Size(1, 1)
    """Sixel aspect ratio."""

    def __init__(
        self,
        *,
        default_color: AColor = TRANSPARENT,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        blitter: Blitter = "half",
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = True,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self.default_color = default_color
        self.alpha = alpha
        self.interpolation = interpolation
        self.texture = np.full((1, 1, 4), default_color, np.uint8)
        self.blitter = blitter  # Property setter will correctly resize texture.

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._alpha

    @alpha.setter
    @bindable
    def alpha(self, alpha: float):
        self._alpha = clamp(float(alpha), 0.0, 1.0)

    @property
    def interpolation(self) -> Interpolation:
        """Interpolation used when gadget is resized."""
        return self._interpolation

    @interpolation.setter
    def interpolation(self, interpolation: Interpolation):
        if interpolation not in Interpolation.__args__:
            raise TypeError(f"{interpolation} is not a valid interpolation type.")
        self._interpolation = interpolation

    @property
    def blitter(self) -> Blitter:
        """Determines how graphics are rendered."""
        return self._blitter

    @blitter.setter
    def blitter(self, blitter: Blitter):
        if blitter not in Blitter.__args__:
            raise TypeError(f"{blitter} is not a valid blitter type.")
        if blitter == "sixel" and not self._sixel_support:
            self._blitter = "half"
            logger.info(
                'Can\'t set blitter to "sixel" without sixel terminal support. '
                'Setting blitter to "half" instead.'
            )
        else:
            self._blitter = blitter
        self.on_size()

    def on_size(self) -> None:
        """Resize texture array."""
        self.texture = resize_texture(
            self.texture, scale_geometry(self._blitter, self.size), self._interpolation
        )

    def on_add(self) -> None:
        """Resize if geometry is incorrect on add."""
        if self._blitter == "sixel" and not Graphics._sixel_support:
            self.blitter = "half"
        elif self.texture.shape[:2] != scale_geometry(self._blitter, self.size):
            self.on_size()
        super().on_add()

    def to_png(self, path: Path) -> None:
        """Write :attr:`texture` to provided path as a `png` image."""
        BGRA = cv2.cvtColor(self.texture, cv2.COLOR_RGBA2BGRA)
        cv2.imwrite(str(path.absolute()), BGRA)

    def clear(self) -> None:
        """Fill texture with default color."""
        self.texture[:] = self.default_color

    def _render(
        self, cells: NDArray[Cell], graphics: NDArray[np.uint8], kind: NDArray[np.uint8]
    ) -> None:
        """Render visible region of gadget."""
        graphics_render(
            cells,
            graphics,
            kind,
            self.absolute_pos,
            self._blitter,
            self._is_transparent,
            self.texture,
            self._alpha,
            self._region,
        )



"""A beam effect."""

import asyncio
from collections.abc import Callable
from random import random, shuffle
from typing import Coroutine, Literal

import numpy as np
from numpy.typing import NDArray

from ...colors import WHITE, Color, gradient
from ..text import Text

HORIZONTAL_BEAM = "▁▁▁▁▁▁▁▁▂▂▂▂▂▂▂▂▃▃▃▃▃▃▃▃▄▄▄▄▄▄▄▄"
VERTICAL_BEAM = "▏\n▏\n▏\n▏\n▎\n▎\n▎\n▎\n▍\n▍\n▍\n▍\n▌\n▌\n▌\n▌"
BEAM_BLUE = Color.from_hex("00d1ff")
BEAM_PURP = Color.from_hex("8a008a")
HORIZONTAL_GRAD = gradient(BEAM_PURP, BEAM_BLUE, WHITE, n=32)
VERTICAL_GRAD = gradient(BEAM_PURP, BEAM_BLUE, WHITE, n=16)


async def beams_effect(text: Text):
    """
    Create a beams effect.

    Parameters
    ----------
    text : Text
        The text gadget which to run the effect.

    Warnings
    --------
    Modifying `text` size while effect is running will break the effect.
    """
    cover = Text(size=text.size)
    cover.canvas[:] = text.canvas
    cover.canvas["fg_color"] = cover.canvas["bg_color"]

    pass_count = np.zeros(text.size, int)
    beams, tweens = _create_beams(pass_count, cover)
    shuffle(tweens)

    cover.add_gadgets(beams)
    text.add_gadget(cover)

    tasks = [asyncio.create_task(_fade_cover(pass_count, cover, text))]
    for tween in tweens:
        tasks.append(asyncio.create_task(tween))
        await asyncio.sleep(0.1 * random())

    await asyncio.gather(*tasks)

    text.remove_gadget(cover)


def _create_on_progress(
    pass_count: NDArray[np.int32],
    cover: Text,
    beam: Text,
    kind: Literal["down", "left", "right", "up"],
) -> Callable[[], None]:
    if kind == "down":
        last_bottom = beam.bottom

        def on_progress(*args):
            nonlocal last_bottom
            if beam.bottom != last_bottom:
                pass_count[last_bottom + 1 : beam.bottom + 1, beam.x] += 1
                cover.canvas[last_bottom + 1 : beam.bottom + 1, beam.x]["fg_color"] = (
                    WHITE
                )
                last_bottom = beam.bottom

    elif kind == "left":
        last_left = beam.left

        def on_progress(*args):
            nonlocal last_left
            if beam.left != last_left:
                left = max(beam.left, 0)
                pass_count[beam.y, left:last_left] += 1
                cover.canvas[beam.y, left:last_left]["fg_color"] = WHITE
                last_left = beam.left

    elif kind == "right":
        last_right = beam.right

        def on_progress(*args):
            nonlocal last_right
            if beam.right != last_right:
                pass_count[beam.y, last_right + 1 : beam.right + 1] += 1
                cover.canvas[beam.y, last_right + 1 : beam.right + 1]["fg_color"] = (
                    WHITE
                )
                last_right = beam.right

    else:
        last_top = beam.top

        def on_progress(*args):
            nonlocal last_top
            if beam.top != last_top:
                top = max(beam.top, 0)
                pass_count[top:last_top, beam.x] += 1
                cover.canvas[top:last_top, beam.x]["fg_color"] = WHITE
                last_top = beam.top

    return on_progress


def _create_beams(
    pass_count: NDArray[np.int32], cover: Text
) -> tuple[list[Text], list[Coroutine]]:
    h, w = cover.size
    beams = []
    tweens = []
    for y in range(h):
        beam = Text(is_transparent=True)
        beam.y = y
        duration = 0.3 + random()
        if round(random()):
            beam.set_text(HORIZONTAL_BEAM)
            beam.canvas["fg_color"] = HORIZONTAL_GRAD
            beam.right = -1
            tween = beam.tween(
                duration=duration,
                on_progress=_create_on_progress(pass_count, cover, beam, "right"),
                left=w,
            )
        else:
            beam.set_text(HORIZONTAL_BEAM[::-1])
            beam.canvas["fg_color"] = HORIZONTAL_GRAD[::-1]
            beam.left = w
            tween = beam.tween(
                duration=duration,
                on_progress=_create_on_progress(pass_count, cover, beam, "left"),
                right=-1,
            )
        beams.append(beam)
        tweens.append(tween)

    for x in range(w):
        beam = Text(is_transparent=True)
        beam.x = x
        duration = 0.3 + random()
        if round(random()):
            beam.set_text(VERTICAL_BEAM)
            beam.canvas["fg_color"][:, 0] = VERTICAL_GRAD
            beam.bottom = -1
            tween = beam.tween(
                duration=duration,
                on_progress=_create_on_progress(pass_count, cover, beam, "down"),
                top=h,
            )
        else:
            beam.set_text(VERTICAL_BEAM[::-1])
            beam.canvas["fg_color"][:, 0] = VERTICAL_GRAD[::-1]
            beam.top = h
            tween = beam.tween(
                duration=duration,
                on_progress=_create_on_progress(pass_count, cover, beam, "up"),
                bottom=-1,
            )
        beams.append(beam)
        tweens.append(tween)
    return beams, tweens


async def _fade_cover(pass_count: NDArray[np.int32], cover: Text, text: Text):
    cover_fg = cover.canvas["fg_color"]
    text_fg = text.canvas["fg_color"].astype(float)
    while not np.array_equal(cover_fg, text_fg):
        passed_mask = pass_count >= 1
        faded = cover_fg * 0.99 + text_fg * 0.01
        cover_fg[passed_mask] = faded.astype(np.uint8)[passed_mask]
        await asyncio.sleep(0)



"""A markdown gadget."""

import asyncio
import re
import webbrowser
from pathlib import Path
from typing import Literal

import cv2
import numpy as np
from mistletoe import Document, block_token, span_token
from mistletoe.base_renderer import BaseRenderer
from numpy.typing import NDArray
from pygments.lexers import get_lexer_by_name
from pygments.style import Style as PygmentsStyle
from pygments.util import ClassNotFound

from ..colors import Neptune, lerp_colors
from ..emojis import EMOJIS
from ..geometry import Point, Size
from ..text_tools import Style
from .behaviors.button_behavior import ButtonBehavior
from .behaviors.themable import Themable
from .gadget import Gadget, PosHint, SizeHint
from .graphics import _BLITTER_GEOMETRY, Blitter, Graphics
from .grid_layout import GridLayout
from .image import Image
from .pane import Pane
from .scroll_view import ScrollView
from .text import Border, Cell, Text, new_cell
from .video import Video

__all__ = ["Markdown", "Point", "Size"]

_TASK_LIST_ITEM_RE = re.compile(r" {,3}\[([xX ])\]\s+(.*)", re.DOTALL)
_BULLETS = "●○◼◻▶▷◆◇"
_MIN_MARKDOWN_WIDTH = 24
_MIN_RENDER_WIDTH = 12

# TODO: Html blocks and spans
# TODO: Wrap Tables
# TODO: Use list prefix for render_width calculation.


def _image_size(texture_h: int, texture_w: int, max_width: int) -> Size:
    """
    Use terminal pixel geometry to scale images.

    If markdown gadgets are created before the terminal pixel geometry has been queried
    then images are scaled by the default pixel geometry of `Size(20, 10)`.
    """
    pixels_per_h, pixels_per_w = _BLITTER_GEOMETRY["sixel"]
    h = int(texture_h / pixels_per_h)
    w = int(texture_w / pixels_per_w)
    if w <= max_width:
        return Size(h, w)

    pct = max_width / w
    return Size(int(h * pct), int(w * pct))


def _is_task_list_item(list_item: block_token.ListItem) -> re.Match | None:
    if list_item.leader != "-":
        return False
    # Match the first raw text item
    current_token = list_item
    while True:
        if isinstance(current_token, span_token.RawText):
            return _TASK_LIST_ITEM_RE.match(current_token.content)
        if current_token.children:
            current_token = current_token.children[0]
        else:
            return None


def _default_cell():
    return new_cell(
        fg_color=Themable.get_color("primary_fg"),
        bg_color=Themable.get_color("primary_bg"),
    )


class BlankLine(block_token.BlockToken):
    pattern = re.compile(r"\s*\n$")

    def __init__(self, _):
        self.children = []

    @classmethod
    def start(cls, line):
        return cls.pattern.match(line)

    @classmethod
    def read(cls, lines):
        return [next(lines)]


# Spaces are tokenized so that `_BatgrlRenderer.render_inner` wraps at words.
class Spaces(span_token.SpanToken):
    pattern = re.compile(r"\b( +)\b")


class EmojiCode(span_token.SpanToken):
    pattern = re.compile(r":([+-]?\w*):")

    def __init__(self, match_obj):
        self.content = EMOJIS.get(match_obj.group(1), match_obj.group(0))


class _BorderedContent(Text):
    def __init__(
        self,
        default_cell: NDArray[Cell],
        border: Border,
        padding: int = 0,
        content: Gadget | None = None,
        bind: bool = False,
    ):
        super().__init__(default_cell=default_cell)

        if content is None:
            self.content = Text(
                pos=(1 + padding, 1 + padding), default_cell=default_cell
            )
        else:
            self.content = content
            content.pos = 1 + padding, 1 + padding

        self.add_gadget(self.content)

        def update():
            h, w = self.content.size
            self.size = h + 2 * (padding + 1), w + 2 * (padding + 1)
            self.add_border(border)

        update()
        if bind:
            self.content.bind("size", update)


class _HasTitle:
    """Title hint behavior for links and images."""

    def __init__(self, title: str, **kwargs):
        super().__init__(**kwargs)
        self.title = title

    def add_gadget(self, gadget):
        super().add_gadget(gadget)
        for child in self.walk():
            if isinstance(child, _HasTitle) and child.title:
                self.title = ""
                break

    def on_add(self):
        for ancestor in self.ancestors():
            if isinstance(ancestor, Markdown):
                self.link_hint = ancestor._link_hint
                self._hint_task = asyncio.create_task(asyncio.sleep(0))  # dummy task
                break
        super().on_add()

    def on_mouse(self, mouse_event):
        if not self.title:
            return super().on_mouse(mouse_event)

        self._myx = mouse_event.pos

        self._hint_task.cancel()
        if self.collides_point(mouse_event.pos):
            self.link_hint.content.set_text(self.title)
            self._hint_task = asyncio.create_task(self.show_hint_soon())
        else:
            self.link_hint.is_enabled = False
        return super().on_mouse(mouse_event)

    async def show_hint_soon(self):
        await asyncio.sleep(1)
        y, x = self.link_hint.parent.to_local(self._myx)
        if y + 1 < self.link_hint.parent.height:
            self.link_hint.pos = y + 1, x
        else:
            self.link_hint.pos = y - 1, x
        self.link_hint.is_enabled = True


class _Link(_HasTitle, ButtonBehavior, Gadget):
    def __init__(self, title, target, content):
        super().__init__(title=title)
        self.target = target
        self.texts = []
        self.graphics = []
        self.add_gadget(content)

        for child in self.walk():
            if isinstance(child, Text):
                style = child.canvas["style"].copy()
                colors = child.canvas[["fg_color", "bg_color"]]
                primary_fg = Themable.get_color("primary_fg")
                primary_bg = Themable.get_color("primary_bg")
                mask = colors == np.array((primary_fg, primary_bg), colors.dtype)
                colors[mask] = (
                    Themable.get_color("markdown_link_fg"),
                    Themable.get_color("markdown_link_bg"),
                )
                self.texts.append((child, style, mask))
            # ! This condition can add to the gadget tree while walking it, careful.
            elif isinstance(child, (_MarkdownImage, _MarkdownGif)):
                if not hasattr(child, "outline"):
                    child.outline = Graphics(size=child.size, is_enabled=False)
                    child.outline.texture[[0, -1]] = (75,) * 4
                    child.outline.texture[:, [0, -1]] = (75,) * 4
                    child.add_gadget(child.outline)
                self.graphics.append(child)
        self.size = content.size

    def on_release(self):
        webbrowser.open(self.target)

    def update_normal(self):
        link_fg = Themable.get_color("markdown_link_fg")
        link_bg = Themable.get_color("markdown_link_bg")
        for text, style, mask in self.texts:
            text.canvas["fg_color"][mask] = link_fg
            text.canvas["bg_color"][mask] = link_bg
            text.canvas["style"] = style
        for graphic in self.graphics:
            graphic.outline.is_enabled = False

    def update_hover(self):
        hover_fg = Themable.get_color("markdown_link_hover_fg")
        hover_bg = Themable.get_color("markdown_link_hover_bg")
        for text, _, mask in self.texts:
            text.canvas["fg_color"][mask] = hover_fg
            text.canvas["bg_color"][mask] = hover_bg
            text.canvas["style"] |= Style.UNDERLINE
        for graphic in self.graphics:
            graphic.outline.is_enabled = True


class _MarkdownImage(_HasTitle, Image):
    def __init__(self, title: str, path: Path, width: int, blitter: Blitter):
        oh, ow, _ = self._otexture.shape
        size = _image_size(oh, ow, width)
        super().__init__(title=title, path=path, blitter=blitter, size=size)


class _MarkdownGif(_HasTitle, Video):
    def __init__(self, title: str, path: Path, width: int, blitter: Blitter):
        super().__init__(title=title, source=path, blitter=blitter)
        oh = self._resource.get(cv2.CAP_PROP_FRAME_HEIGHT)
        ow = self._resource.get(cv2.CAP_PROP_FRAME_WIDTH)
        self.size = _image_size(oh, ow, width)

    def on_add(self):
        super().on_add()
        self.play()

    def on_remove(self):
        self.stop()
        super().on_remove()


class _TextImage(_HasTitle, Gadget):
    def __init__(self, title, content):
        super().__init__(title=title)
        self.add_gadget(content)
        self.size = content.size


class _BlockCode(Text):
    def __init__(self):
        super().__init__(default_cell=_default_cell())


class _List(GridLayout):
    def __init__(self, items, prefix: str | int):
        super().__init__(
            grid_rows=len(items),
            grid_columns=2,
            horizontal_spacing=1,
            is_transparent=True,
        )
        if isinstance(prefix, int):  # ordered list
            for i, item in enumerate(items, start=prefix):
                number = Text(default_cell=_default_cell())
                number.set_text(f"{i}.")
                self.add_gadgets(number, item)
        else:  # unordered list
            for item in items:
                bullet = Text(default_cell=_default_cell())
                bullet.set_text(getattr(item, "check", prefix))
                self.add_gadgets(bullet, item)

        self.size = self.min_grid_size


class _Quote(Pane):
    def __init__(self, content: Gadget, depth):
        super().__init__()
        quote_fg = Themable.get_color("markdown_quote_fg")
        quote_bg = Themable.get_color("markdown_quote_bg")
        margin = Text(
            default_cell=new_cell(ord=ord("▎"), fg_color=quote_fg, bg_color=quote_bg),
            size=(content.height, 1),
        )
        self.depth = depth
        self.bg_color = lerp_colors(quote_fg, quote_bg, 1 / (0.1 * depth + 1))
        self.grid = GridLayout(grid_rows=1, grid_columns=2, is_transparent=True)
        self.grid.add_gadgets(margin, content)
        self.add_gadget(self.grid)
        self.size = self.grid.size = self.grid.min_grid_size
        self.color_content()

    def color_content(self):
        self._is_colored = True
        for child in self.walk():
            if hasattr(child, "_is_colored"):
                continue

            child._is_colored = True
            if isinstance(child, _BlockCode):
                child.canvas["bg_color"] = Themable.get_color(
                    "markdown_quote_block_code_bg"
                )
            elif isinstance(child, Text):
                child.canvas["bg_color"] = self.bg_color
            elif isinstance(child, Pane):
                child.bg_color = self.bg_color


class _BatgrlRenderer(BaseRenderer):
    list_depth: int = 0
    quote_depth: int = 0
    last_token: span_token.SpanToken | block_token.BlockToken | None = None

    def __init__(self, width, syntax_highlighting_style, blitter):
        super().__init__(BlankLine, Spaces, EmojiCode)
        block_token.remove_token(block_token.Footnote)
        self.width = max(width, _MIN_MARKDOWN_WIDTH)
        self.syntax_highlighting_style = syntax_highlighting_style
        self.blitter = blitter
        self.render_map["SetextHeading"] = self.render_setext_heading
        self.render_map["CodeFence"] = self.render_block_code

    @property
    def render_width(self):
        return max(
            self.width - 3 * self.list_depth - 2 * self.quote_depth, _MIN_RENDER_WIDTH
        )

    def render(self, token: span_token.SpanToken | block_token.BlockToken) -> Gadget:
        if isinstance(self.last_token, BlankLine) and isinstance(
            token, block_token.Paragraph
        ):
            self.last_rendered.skip_blank_line = True
        self.last_rendered = super().render(token)
        self.last_token = token
        return self.last_rendered

    def render_blank_line(self, token: BlankLine) -> Gadget:
        return Gadget(size=(1, 1), is_transparent=True)

    def render_spaces(self, token: Spaces) -> Literal[" "]:
        return " "

    def render_raw_text(self, token: span_token.RawText) -> Text:
        text = Text(default_cell=_default_cell())
        text.set_text(token.content)
        width = self.render_width
        if text.width > width:
            new_lines, line_end = divmod(text.width, width)
            text.height += new_lines
            for i in range(1, new_lines):
                text.canvas[i, :width] = text.canvas[0, i * width : (i + 1) * width]
            last_line = new_lines * width
            text.canvas[new_lines, :line_end] = text.canvas[
                0, last_line : last_line + line_end
            ]
            text.line_end = line_end
            text.width = width
        return text

    def render_emoji_code(self, token: EmojiCode) -> Text:
        return self.render_raw_text(token)

    def render_inner(self, token: span_token.SpanToken) -> Text:
        width = self.render_width
        text = Text(size=(1, 1), default_cell=_default_cell())
        current_line_end = 0
        current_line_height = 1
        current_line_gadgets = []
        for content in map(self.render, token.children):
            if content == "\n" or content == " " and current_line_end + 1 == text.width:
                text.height += 1
                current_line_end = 0
                current_line_height = 1
                current_line_gadgets = []
                continue

            if content == " ":
                current_line_end += 1
                continue

            new_line_end = current_line_end + content.width
            if new_line_end > text.width:
                if new_line_end <= width:
                    text.width = new_line_end
                else:
                    text.height += 1
                    new_line_end = content.width
                    if text.width < new_line_end:
                        text.width = new_line_end
                    current_line_end = 0
                    current_line_height = 1
                    current_line_gadgets = []

            if isinstance(content, Text):
                if content.height > 1:
                    text.height += content.height - 1

                text.canvas[-content.height :, current_line_end:new_line_end] = (
                    content.canvas
                )
                current_line_end = getattr(content, "line_end", new_line_end)
                continue

            if content.height > current_line_height:
                height_dif = content.height - current_line_height
                current_line_height = content.height
                text.height += height_dif
                # Move text on current line down
                old = text.canvas[:-height_dif, :current_line_end]
                text.canvas[height_dif:, :current_line_end] = old
                old[:] = text.default_cell
                # Move links and images on current line down.
                for gadget in current_line_gadgets:
                    gadget.y += height_dif

            content.pos = text.height - content.height, current_line_end
            current_line_end = new_line_end
            current_line_gadgets.append(content)
            text.add_gadget(content)

        return text

    def render_strong(self, token: span_token.Strong) -> Text:
        text = self.render_inner(token)
        if text.height > 1:
            text.canvas[:-1]["style"] |= Style.BOLD
            text.canvas[-1, : text.line_end]["style"] |= Style.BOLD
        else:
            text.canvas["style"] |= Style.BOLD
        return text

    def render_emphasis(self, token: span_token.Emphasis) -> Text:
        text = self.render_inner(token)
        if text.height > 1:
            text.canvas[:-1]["style"] |= Style.ITALIC
            text.canvas[-1, : text.line_end]["style"] |= Style.ITALIC
        else:
            text.canvas["style"] |= Style.ITALIC
        return text

    def render_strikethrough(self, token: span_token.Strikethrough) -> Text:
        text = self.render_inner(token)
        if text.height > 1:
            text.canvas[:-1]["style"] = Style.STRIKETHROUGH
            text.canvas[-1, : text.line_end]["style"] = Style.STRIKETHROUGH
        else:
            text.canvas["style"] = Style.STRIKETHROUGH
        return text

    def render_inline_code(self, token: span_token.InlineCode) -> Text:
        text = self.render_raw_text(token.children[0])
        inline_fg = Themable.get_color("markdown_inline_code_fg")
        inline_bg = Themable.get_color("markdown_inline_code_bg")
        if text.height > 1:
            text.canvas["fg_color"][:-1] = inline_fg
            text.canvas["bg_color"][:-1] = inline_bg
            text.canvas["fg_color"][-1, : text.line_end] = inline_fg
            text.canvas["bg_color"][-1, : text.line_end] = inline_bg
        else:
            text.canvas["fg_color"] = inline_fg
            text.canvas["bg_color"] = inline_bg
        return text

    def render_image(
        self, token: span_token.Image
    ) -> _MarkdownGif | _MarkdownImage | _TextImage:
        path = Path(token.src)
        if path.exists():
            if path.suffix == ".gif":
                return _MarkdownGif(
                    path=path,
                    title=token.title,
                    width=self.render_width,
                    blitter=self.blitter,
                )
            return _MarkdownImage(
                path=path,
                title=token.title,
                width=self.render_width,
                blitter=self.blitter,
            )
        token.children.insert(0, span_token.RawText("🖼️  "))
        content = self.render_inner(token)
        content.canvas["fg_color"] = Themable.get_color("markdown_image_fg")
        content.canvas["bg_color"] = Themable.get_color("markdown_image_bg")
        return _TextImage(title=token.title, content=content)

    def render_link(self, token: span_token.Link) -> _Link:
        line = _Link(
            target=token.target, title=token.title, content=self.render_inner(token)
        )
        return line

    def render_auto_link(self, token: span_token.AutoLink) -> _Link:
        target = f"mailto:{token.target}" if token.mailto else token.target
        return _Link(
            target=target, title=token.target, content=self.render_inner(token)
        )

    def render_escape_sequence(self, token: span_token.EscapeSequence) -> Text:
        text = Text(default_cell=_default_cell())
        text.set_text(f"\\{token.children[0].content}")
        return text

    def render_heading(self, token: block_token.Heading) -> Text:
        text = self.render_inner(token)
        if token.level % 2 == 1:
            text.canvas["style"] |= Style.BOLD
        if token.level < 5:
            primary_fg = Themable.get_color("primary_fg")
            primary_bg = Themable.get_color("primary_bg")
            header_bg = Themable.get_color("markdown_header_bg")
            text.canvas["bg_color"] = header_bg
            header = _BorderedContent(
                default_cell=new_cell(fg_color=primary_fg, bg_color=header_bg),
                border="mcgugan_wide",
                padding=int(token.level < 3),
                content=text,
            )
            header.canvas["bg_color"][[0, -1]] = primary_bg
            return header
        text.height += 1
        text.chars[-1] = "▔"
        return text

    def render_setext_heading(self, token: block_token.SetextHeading) -> Text:
        text = self.render_inner(token)
        if token.level == 1:
            text.canvas["style"] |= Style.BOLD
        text.height += 1
        text.width = max(text.width, self.render_width)
        text.chars[-1, :] = "━" if token.level == 1 else "─"
        return text

    def render_quote(self, token: block_token.Quote) -> _Quote:
        self.quote_depth += 1
        items = list(map(self.render, token.children))
        self.quote_depth -= 1

        if len(items) == 1:
            quote = _Quote(items[0], self.quote_depth)
        else:
            grid = GridLayout(grid_rows=len(items), is_transparent=True)
            grid.add_gadgets(items)
            grid.size = grid.min_grid_size
            quote = _Quote(grid, self.quote_depth)

        quote.width = max(quote.width, self.render_width)
        return quote

    def render_paragraph(self, token: block_token.Paragraph) -> Text:
        return self.render_inner(token)

    def render_block_code(
        self, token: block_token.BlockCode | block_token.CodeFence
    ) -> _BlockCode:
        text = _BlockCode()
        text.set_text(token.content.rstrip())
        text.width = max(text.width, self.render_width)
        try:
            lexer = get_lexer_by_name(token.language)
        except ClassNotFound:
            lexer = None
        text.add_syntax_highlighting(lexer=lexer, style=self.syntax_highlighting_style)
        text.canvas["bg_color"] = Themable.get_color("markdown_block_code_bg")
        return text

    def render_list(self, token: block_token.List) -> _List:
        prefix: str | int
        if token.start is None:
            nbullets = len(_BULLETS)
            prefix = _BULLETS[self.list_depth % nbullets]
        else:
            prefix = token.start

        self.list_depth += 1
        checks = list(map(_is_task_list_item, token.children))
        if all(checks):
            items = [
                self.render_task_list_item(child, match)
                for child, match in zip(token.children, checks)
            ]
        else:
            items = [self.render_list_item(child) for child in token.children]
        self.list_depth -= 1
        return _List(items, prefix)

    def render_task_list_item(
        self, token: block_token.ListItem, match: re.Match
    ) -> Gadget:
        current_token = token
        while True:
            if isinstance(current_token, span_token.RawText):
                current_token.content = match[2]
                list_item = self.render_list_item(token)
                list_item.check = "🟩" if match[1] == " " else "❎"
                return list_item
            current_token = current_token.children[0]

    def render_list_item(self, token: block_token.ListItem) -> Gadget:
        blocks = list(map(self.render, token.children))

        if len(blocks) == 1:
            return blocks[0]
        list_item = GridLayout(
            grid_columns=1, grid_rows=len(blocks), is_transparent=True
        )
        list_item.add_gadgets(blocks)
        list_item.size = list_item.min_grid_size
        return list_item

    def render_table(self, token: block_token.Table) -> Text:
        header = token.header.children if hasattr(token, "header") else []
        rows = [header, *(row.children for row in token.children)]
        rendered_rows = [[self.render_inner(cell) for cell in row] for row in rows]

        MIN_COLUMN_WIDTH = 3
        row_heights = []
        column_widths = [MIN_COLUMN_WIDTH] * len(header)
        alignments = token.column_align.copy()
        for row in rendered_rows:
            row_height = 1
            column_widths.extend([MIN_COLUMN_WIDTH] * (len(row) - len(column_widths)))
            alignments.extend([None] * (len(row) - len(alignments)))
            for i, cell in enumerate(row):
                if cell.width > column_widths[i]:
                    column_widths[i] = cell.width
                if cell.height > row_height:
                    row_height = cell.height
            row_heights.append(row_height)

        for row, row_height in zip(rendered_rows, row_heights):
            for cell, column_width, alignment in zip(row, column_widths, alignments):
                if cell.height < row_height:
                    cell.height = row_height

                if column_width == cell.width:
                    continue

                diff = column_width - cell.width
                cell.width += diff
                if alignment == 0:  # centered
                    offset_left = diff // 2
                    offset_right = diff - offset_left
                    cell.canvas[:, offset_left:-offset_right] = cell.canvas[:, :-diff]
                    cell.canvas[:, :offset_left] = cell.default_cell
                elif alignment == 1:  # right-aligned
                    cell.canvas[:, diff:] = cell.canvas[:, :-diff]
                    cell.canvas[:, :diff] = cell.default_cell

        OUTER_PAD = 1
        INNER_PAD = 2
        table = Text(
            size=(
                sum(row_heights) + len(rows) - 1,
                sum(column_widths)
                + OUTER_PAD * 2
                + INNER_PAD * (len(column_widths) - 1),
            ),
            default_cell=_default_cell(),
        )
        y = 0
        for i, row in enumerate(rendered_rows):
            x = OUTER_PAD
            for cell in row:
                h, w = cell.size
                table.canvas[y : y + h, x : x + w] = cell.canvas
                children = cell.children.copy()
                for child in children:
                    cell.remove_gadget(child)
                    cy, cx = child.pos
                    child.pos = cy + y, cx + x
                    table.add_gadget(child)
                x += w + INNER_PAD
            y += h
            if i == 0:
                table.canvas["style"][:y] |= Style.BOLD
                table.chars[y] = "━"
            elif i < len(rows) - 1:
                table.chars[y] = "─"
            y += 1

        return table

    def render_thematic_break(self, token: block_token.ThematicBreak) -> Text:
        cell = _default_cell()
        cell["ord"] = ord("─")
        return Text(size=(1, self.render_width), default_cell=cell)

    def render_line_break(self, token: span_token.LineBreak) -> Literal[" ", "\n"]:
        return " " if token.soft else "\n"

    def render_document(self, token: block_token.Document) -> GridLayout:
        # All blocks need to be rendered before skipped blank lines can be determined.
        rendered = [self.render(child) for child in token.children]
        blocks = [block for block in rendered if not hasattr(block, "skip_blank_line")]
        grid = GridLayout(grid_rows=len(blocks), is_transparent=True)
        grid.add_gadgets(blocks)
        grid.size = grid.min_grid_size
        grid.footnotes = token.footnotes
        return grid


class Markdown(Themable, Gadget):
    r"""
    A markdown gadget.

    Parameters
    ----------
    markdown : str
        The markdown string.
    syntax_highlighting_style : pygments.style.Style, default: Neptune
        The syntax highlighting style for code blocks.
    blitter : Blitter, default: "half"
        Determines how images are rendered.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    markdown : str
        The markdown string.
    syntax_highlighting_style : pygments.style.Style
        The syntax highlighting style for code blocks.
    blitter : Blitter, default: "half"
        Determines how images are rendered.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    get_color()
        Get a color by name from the current color theme.
    update_theme()
        Paint the gadget with current theme.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        markdown: str,
        syntax_highlighting_style: PygmentsStyle = Neptune,
        blitter: Blitter = "half",
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self._scroll_view = ScrollView(
            size_hint={
                "height_hint": 1.0,
                "width_hint": 1.0,
                "min_width": _MIN_MARKDOWN_WIDTH,
            },
            dynamic_bars=True,
        )
        title_fg = Themable.get_color("markdown_title_fg")
        title_bg = Themable.get_color("markdown_title_bg")
        self._link_hint = _BorderedContent(
            default_cell=new_cell(fg_color=title_fg, bg_color=title_bg),
            border="outer",
            bind=True,
        )
        self._link_hint.is_enabled = False
        self.add_gadgets(self._scroll_view, self._link_hint)
        self.markdown: str = markdown
        self.syntax_highlighting_style: PygmentsStyle = syntax_highlighting_style
        """The syntax highlighting style for code blocks."""
        self.blitter = blitter
        """Determines how images are rendered."""

    @property
    def markdown(self) -> str:
        """The markdown string."""
        return self._markdown

    @markdown.setter
    def markdown(self, markdown: str):
        self._markdown = markdown
        self._build_markdown()

    @property
    def syntax_highlighting_style(self) -> PygmentsStyle:
        """The syntax highlighting style for code blocks."""
        return self._style

    @syntax_highlighting_style.setter
    def syntax_highlighting_style(self, syntax_highlighting_style: PygmentsStyle):
        self._style = syntax_highlighting_style
        self._build_markdown()

    @property
    def blitter(self) -> Blitter:
        """Determines how images are rendered."""
        return self._blitter

    @blitter.setter
    def blitter(self, blitter: Blitter):
        self._blitter = blitter
        for child in self.walk():
            if isinstance(child, Graphics):
                child.blitter = blitter

    def _build_markdown(self):
        if not self.root:
            return

        with _BatgrlRenderer(
            self._scroll_view.port_width, self.syntax_highlighting_style, self.blitter
        ) as renderer:
            rendered = renderer.render(Document(self.markdown))

        if self._scroll_view.view is not None:
            self._scroll_view.view.destroy()
        self._scroll_view.view = rendered

    def update_theme(self):
        """Paint the gadget with current theme."""
        title_fg = Themable.get_color("markdown_title_fg")
        title_bg = Themable.get_color("markdown_title_bg")
        title_cell = new_cell(fg_color=title_fg, bg_color=title_bg)
        self._link_hint.default_cell = title_cell
        self._link_hint.content.default_cell = title_cell
        self._link_hint.canvas[:] = title_cell
        self._link_hint.content.canvas[:] = title_cell
        self._build_markdown()

    def on_size(self):
        """Rebuild markdown on resize."""
        super().on_size()
        self._build_markdown()

    def on_add(self):
        """Build markdown on add."""
        super().on_add()
        self._build_markdown()



"""Root gadget."""

from __future__ import annotations

from threading import RLock
from typing import TYPE_CHECKING, Final, Literal, Self

import numpy as np
from numpy.typing import NDArray

from ..colors import BLACK, Color
from ..text_tools import Cell, new_cell
from .gadget import Gadget, Point, Region, Size, _GadgetList
from .graphics import Graphics, scale_geometry

if TYPE_CHECKING:
    from ..app import App


class _Root(Gadget):
    """
    Root gadget of the gadget tree.

    Instantiated only by :class:`batgrl.app.App`.
    """

    def __init__(self, app: App, size: Size):
        self.children = _GadgetList()

        self._render_lock = RLock()
        """Lock held during rendering to prevent errors related to invalid geometry."""
        self._app = app
        """The running app."""
        self._cell: Cell = new_cell()
        """Default cell of root canvas."""
        self._bg_color = BLACK
        """Background color of the app."""
        self._regions_valid = False
        """Whether all regions in gadget tree are valid."""
        self._pos: Final = Point(0, 0)
        """Position of root gadget."""
        self._size = size
        """Size of root gadget."""

        # Following attributes set in `on_size()`:
        self._resized: bool
        """Whether terminal has resized since last render."""
        self.cells: NDArray[Cell]
        """Current rendering of gadget tree."""
        self.graphics: NDArray[np.uint8] = np.empty((0, 0, 4), np.uint8)
        """Current graphics rendering."""
        self._sgraphics: NDArray[np.uint8] = np.empty((0, 0, 4), np.uint8)
        """Used by renderer to scale graphics from 0-255 to 0-100 for sixel."""
        self.kind: NDArray[np.uint8]
        """Whether a cell should use canvas, graphics or both."""
        self._widths: NDArray[np.int32]
        """Column widths of characters in canvas."""
        self._last_cells: NDArray[Cell]
        """Previous rendering of gadget tree."""
        self._last_graphics: NDArray[np.uint8] = np.empty((0, 0, 4), np.uint8)
        """Previous graphics rendering."""
        self._last_kind: NDArray[np.uint8]
        """Previous kind."""

    def on_size(self):
        """Remake buffers and set ``_resized`` flag on resize."""
        self._resized = True
        self.cells = np.full(self._size, self._cell)
        self._last_cells = self.cells.copy()
        self._widths = np.zeros(self._size, dtype=np.int32)
        self.kind = np.zeros(self._size, np.uint8)
        self._last_kind = self.kind.copy()

        if Graphics._sixel_support:
            h, w = scale_geometry("sixel", self.size)
            self.graphics = np.full((h, w, 4), (*self._bg_color, 0), np.uint8)
            self._last_graphics = self.graphics.copy()
            self._sgraphics = np.zeros((h, w, 4), np.uint8)

    @property
    def bg_color(self) -> Color:
        """Background color of the app."""
        return self._bg_color

    @bg_color.setter
    def bg_color(self, bg_color: Color):
        self._bg_color = self._cell["bg_color"] = bg_color

    @property
    def pos(self) -> Point:
        return self._pos

    @property
    def is_transparent(self) -> Literal[False]:
        return False

    @property
    def is_visible(self) -> Literal[True]:
        return True

    @property
    def is_enabled(self) -> Literal[True]:
        return True

    @property
    def parent(self) -> Literal[None]:
        return None

    @property
    def root(self) -> Self:
        return self

    @property
    def app(self) -> App:
        """The running app."""
        return self._app

    def _set_regions(self) -> None:
        """Recompute all gadget regions."""
        self._region = Region.from_rect(self._pos, self.size)

        for child in self.walk():
            child._region = (
                child.parent._region & Region.from_rect(child.absolute_pos, child.size)
                if child._is_enabled and child._is_visible
                else Region()
            )

        for child in self.walk_reverse():
            if child._is_enabled and child._is_visible:
                child._region &= self._region
                if not child._is_transparent:
                    self._region -= child._region

        self._regions_valid = True
        self._resized = False

    def _render(self):
        """Render gadget tree into :attr:``canvas``."""
        with self._render_lock:
            if not self._regions_valid:
                self._set_regions()

            self.cells, self._last_cells = self._last_cells, self.cells
            self.graphics, self._last_graphics = self._last_graphics, self.graphics
            self.kind, self._last_kind = self._last_kind, self.kind

            self.cells[:] = self._cell
            self.graphics[:] = (*self.bg_color, 0)
            self.kind[:] = 0

            for child in self.walk():
                if not child._is_enabled or not child._is_visible:
                    continue

                child._render(self.cells, self.graphics, self.kind)



"""A textbox gadget for single-line editable text."""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import astuple

from numpy.typing import NDArray
from ugrapheme import grapheme_len, graphemes
from uwcwidth import wcswidth

from ..geometry import rect_slice
from ..logging import get_logger
from ..terminal.events import KeyEvent, MouseButton, MouseEvent, PasteEvent
from ..text_tools import canvas_as_text, egc_chr, is_word_char
from .behaviors.focusable import Focusable
from .behaviors.grabbable import Grabbable
from .behaviors.themable import Themable
from .cursor import Cursor
from .gadget import Cell, Gadget, Point, PosHint, Region, Size, SizeHint
from .text import Text

__all__ = ["Textbox", "Point", "Size"]

logger = get_logger(__name__)


class _Box(Text):
    def _render(self, canvas, graphics, kind):
        super()._render(canvas, graphics, kind)
        textbox: Textbox = self.parent
        if textbox.hide_input:
            hider_rect = Region.from_rect(self.absolute_pos, (1, textbox._line_width))
            hider_region = self._region & hider_rect
            for pos, size in hider_region.rects():
                # FIXME: Allow for egcs for textbox.hide_char
                canvas["ord"][rect_slice(pos, size)] = ord(textbox.hide_char)


class Textbox(Themable, Focusable, Grabbable, Gadget):
    r"""
    A textbox gadget for single-line editable text.

    Supports pasting, mouse selection, and cursor navigation.

    Parameters
    ----------
    enter_callback : Callable[[Textbox], None] | None, default: None
        Called when textbox has focus and `enter` is pressed.
    placeholder : str, default: ""
        Placeholder text for textbox.
    hide_input : bool, default: False
        Whether input is hidden with :attr:`hide_char`.
    hide_char : str, default: "*"
        Character to hide input when :attr:`hide_input` is true.
    max_chars : int | None, default: None
        Maximum allowed number of characters in textbox.
    alpha : float, default: 1.0
        Transparency of gadget.
    is_grabbable : bool, default: True
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool, default: False
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton, default: "left"
        Mouse button used for grabbing.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    enter_callback : Callable[[Textbox], None] | None
        Called when textbox has focus and `enter` is pressed.
    placeholder : str
        Placeholder text for textbox.
    hide_input : bool
        Whether input is hidden with :attr:`hide_char`.
    hide_char : str
        Character to hide input when :attr:`hide_input` is true.
    max_chars : int | None
        Maximum allowed number of characters in textbox.
    alpha : float
        Transparency of gadget.
    text : str
        The textbox's text.
    cursor : int
        The cursor column.
    is_selecting : bool
        Whether there is a selection.
    has_nonempty_selection : bool
        Whether selection is non-empty.
    is_focused : bool
        Whether gadget has focus.
    any_focused : bool
        Whether any gadget has focus.
    is_grabbable : bool
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton
        Mouse button used for grabbing.
    is_grabbed : bool
        Whether gadget is grabbed.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    undo()
        Undo previous edit.
    redo()
        Redo previous undo.
    select()
        Start a new selection at cursor if none.
    unselect()
        Unselect current selection.
    delete_selection()
        Delete current selection.
    move_cursor_left(n)
        Move cursor left `n` characters.
    move_cursor_right(n)
        Move cursor right `n` characters.
    move_word_left()
        Move cursor a word left.
    move_word_right()
        Move cursor a word right.
    get_color()
        Get a color by name from the current color theme.
    update_theme()
        Paint the gadget with current theme.
    focus()
        Focus gadget.
    blur()
        Un-focus gadget.
    focus_next()
        Focus next focusable gadget.
    focus_previous()
        Focus previous focusable gadget.
    on_focus()
        Update gadget when it gains focus.
    on_blur()
        Update gadget when it loses focus.
    grab(mouse_event)
        Grab the gadget.
    ungrab(mouse_event)
        Ungrab the gadget.
    grab_update(mouse_event)
        Update gadget with incoming mouse events while grabbed.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        enter_callback: Callable[[Textbox], None] | None = None,
        placeholder: str = "",
        hide_input: bool = False,
        hide_char: NDArray[Cell] | str = "*",
        max_chars: int | None = None,
        is_grabbable: bool = True,
        ptf_on_grab: bool = False,
        mouse_button: MouseButton = "left",
        alpha: float = 1.0,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self._placeholder_gadget = Text(alpha=0.0, is_transparent=is_transparent)
        self._placeholder_gadget.set_text(placeholder)
        self._cursor = Cursor(reverse=False)
        self._box = _Box(size=size, is_transparent=is_transparent)
        super().__init__(
            is_grabbable=is_grabbable,
            ptf_on_grab=ptf_on_grab,
            mouse_button=mouse_button,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self._selection_start = self._selection_end = None
        self._line_width = 0
        self._undo_stack = []
        self._redo_stack = []
        self._undo_buffer = []
        self._undo_buffer_type = "add"

        self._box.add_gadgets(self._placeholder_gadget, self._cursor)
        self.add_gadgets(self._box)

        self.enter_callback = enter_callback
        """Called when textbox has focus and `enter` is pressed."""
        self.placeholder = placeholder
        """Placeholder text for textbox."""
        self.hide_input = hide_input
        """Whether input is hidden with :attr:`hide_char`."""
        self.hide_char = hide_char
        """Character to hide input if :attr:`hide_input` is true."""
        self.max_chars = max_chars
        """Maximum allowed number of characters in textbox."""
        self.alpha = alpha

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._box.alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._box.alpha = alpha

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        self._box.is_transparent = self.is_transparent
        self._placeholder_gadget.is_transparent = self.is_transparent

    @property
    def hide_char(self) -> str:
        """Character to hide input when :attr:`hide_input` is true."""
        return self._hide_char

    @hide_char.setter
    def hide_char(self, char: str):
        self._hide_char = char[:1] or "*"

    def update_theme(self):
        """Paint the gadget with current theme."""
        primary_fg = self.get_color("textbox_primary_fg")
        primary_bg = self.get_color("textbox_primary_bg")
        self._box.canvas["fg_color"] = self._box.default_fg_color = primary_fg
        self._box.canvas["bg_color"] = self._box.default_bg_color = primary_bg
        self._cursor.fg_color = primary_bg
        self._cursor.bg_color = primary_fg

        placeholder_fg = self.get_color("textbox_placeholder_fg")
        placeholder_bg = self.get_color("textbox_placeholder_bg")
        self._placeholder_gadget.default_fg_color = placeholder_fg
        self._placeholder_gadget.default_bg_color = placeholder_bg
        self._placeholder_gadget.canvas["fg_color"] = placeholder_fg
        self._placeholder_gadget.canvas["bg_color"] = placeholder_bg

        self._highlight_selection()

    def on_size(self):
        """Resize and reposition children on resize."""
        self._box.width = max(self.width, self._line_width + 1)
        if self._box.x + self._line_width < self.width:
            self._box.x = min(0, self.width - self._line_width)
        self.cursor = self.cursor

    def on_focus(self):
        """Show cursor and select all text on focus."""
        self._cursor.is_enabled = True
        if self._line_width > 0:
            self._ctrl_a()

    def on_blur(self):
        """Hide cursor on blur."""
        self._cursor.is_enabled = False
        self.unselect()
        self.cursor = self.cursor

    @property
    def placeholder(self) -> str:
        """Placeholder text for textbox."""
        return self._placeholder

    @placeholder.setter
    def placeholder(self, placeholder: str):
        self._placeholder = placeholder
        self._placeholder_gadget.set_text(placeholder)
        self._placeholder_gadget.is_enabled = self._line_width == 0 and bool(
            placeholder
        )

    def _move_undo_buffer_to_stack(self, buffer_type=None):
        self._undo_buffer_type = buffer_type
        if self._undo_buffer:
            self._undo_stack.append(self._undo_buffer)
            self._undo_buffer = []
            self._redo_stack.clear()

    def undo(self):
        """Undo previous edit."""
        self._move_undo_buffer_to_stack()
        if self._undo_stack:
            redo = []
            for func, args, selection_start, selection_end, cursor in reversed(
                self._undo_stack.pop()
            ):
                redo.append(func(*args))
                self._selection_start = selection_start
                self._selection_end = selection_end
                self.cursor = cursor
            self._redo_stack.append(redo)

    def redo(self):
        """Redo previous undo."""
        if self._redo_stack and not self._undo_buffer:
            undo = []
            for func, args, selection_start, selection_end, cursor in reversed(
                self._redo_stack.pop()
            ):
                undo.append(func(*args))
                self._selection_start = selection_start
                self._selection_end = selection_end
                self.cursor = cursor
            self._undo_stack.append(undo)

    @property
    def text(self) -> str:
        """The textbox's text."""
        return canvas_as_text(self._box.canvas[0, : self._line_width])

    @text.setter
    def text(self, text: str):
        text = text.replace("\n", " ")[: self.max_chars]
        self.unselect()
        self._del_text(0, self._line_width)
        self._add_text(0, text)
        self._redo_stack.clear()
        self._undo_stack.clear()
        self._undo_buffer.clear()
        self.cursor = self._line_width

    @property
    def cursor(self) -> int:
        """The cursor column."""
        return self._cursor.x

    @cursor.setter
    def cursor(self, cursor: int):
        """After setting cursor position, move textbox so that cursor is visible."""
        self._cursor.x = cursor
        self._placeholder_gadget.is_enabled = self._line_width == 0 and bool(
            self._placeholder
        )

        max_x = self.width - 1
        rel_x = cursor + self._box.x
        if rel_x > max_x:
            self._box.x -= rel_x - max_x
        elif rel_x < 0:
            self._box.x -= rel_x

        if self.is_selecting:
            self._selection_end = self.cursor

        self._highlight_selection()

    def _highlight_selection(self):
        fg = self._box.canvas["fg_color"]
        bg = self._box.canvas["bg_color"]
        fg[:] = self._box.default_fg_color
        bg[:] = self._box.default_bg_color

        if self._selection_start != self._selection_end:
            if self._selection_start > self._selection_end:
                start = self._selection_end
                end = self._selection_start
            else:
                start = self._selection_start
                end = self._selection_end

            fg[0, start:end] = self.get_color("textbox_selection_highlight_fg")
            bg[0, start:end] = self.get_color("textbox_selection_highlight_bg")

    @property
    def is_selecting(self) -> bool:
        """Whether there is a selection."""
        return self._selection_start is not None and self._selection_end is not None

    @property
    def has_nonempty_selection(self) -> bool:
        """Whether selection is non-empty."""
        return self.is_selecting and self._selection_start != self._selection_end

    def select(self):
        """Start a new selection at cursor if none."""
        if not self.is_selecting:
            self._selection_start = self._selection_end = self.cursor

    def unselect(self):
        """Unselect current selection."""
        self._selection_start = self._selection_end = None

    def delete_selection(self):
        """Delete current selection."""
        if self.has_nonempty_selection:
            return self._del_text(self._selection_start, self._selection_end)

    def _del_text(self, start: int, end: int):
        if start > end:
            start, end = end, start

        if end > self._line_width:
            logger.debug("End of delete greater than line length.")
            end = self._line_width

        contents = canvas_as_text(self._box.canvas[0, start:end])
        selection_start = self._selection_start
        selection_end = self._selection_end
        cursor = self.cursor

        len_end = self._line_width - end
        self._line_width = start + len_end

        box = self._box
        box.canvas[0, start : self._line_width] = box.canvas[0, end : end + len_end]
        box.canvas[0, self._line_width :] = box.default_cell

        self.unselect()
        self.cursor = start

        return self._add_text, [start, contents], selection_start, selection_end, cursor

    def _add_text(self, x: int, text: str):
        selection_start = self._selection_start
        selection_end = self._selection_end
        cursor = self.cursor

        box_text = graphemes(
            canvas_as_text(self._box.canvas[0, :x])
            + text.replace("\n", " ")
            + canvas_as_text(self._box.canvas[0, x : self._line_width])
        )[: self.max_chars]

        box_width = self._line_width = wcswidth(box_text)
        if box_width >= self._box.width:
            self._box.width = box_width + 1

        self._box.add_str(box_text)
        self._box.canvas[0, box_width:] = self._box.default_cell

        self.cursor = min(box_width, x + wcswidth(text))
        return self._del_text, [x, self.cursor], selection_start, selection_end, cursor

    def move_cursor_left(self, n: int = 1):
        """Move cursor left `n` characters."""
        text_before_cursor = canvas_as_text(self._box.canvas[0, : self.cursor])
        egcs = graphemes(text_before_cursor)
        if n <= len(egcs):
            self.cursor = wcswidth(egcs[:-n])
        else:
            self.cursor = 0

    def move_cursor_right(self, n: int = 1):
        """Move cursor right `n` characters."""
        text_after_cursor = canvas_as_text(
            self._box.canvas[0, self.cursor : self._line_width]
        )
        egcs = graphemes(text_after_cursor)
        if n <= len(egcs):
            self.cursor += wcswidth(egcs[:n])
        else:
            self.cursor = self._line_width

    def move_word_left(self):
        """Move cursor a word left."""
        last_x = self.cursor
        first_char_found = False
        while True:
            self.move_cursor_left()
            if self.cursor == last_x:
                break

            last_x = self.cursor

            current_char = egc_chr(self._box.canvas[0, self.cursor]["ord"])
            if not first_char_found:
                if not current_char.isspace():
                    first_char_found = True
                    char_is_word_char = is_word_char(current_char)
            elif current_char.isspace() or char_is_word_char != is_word_char(
                current_char
            ):
                self.move_cursor_right()
                break

    def move_word_right(self):
        """Move cursor a word right."""
        last_x = self.cursor
        first_char_found = False
        while True:
            self.move_cursor_right()
            if self.cursor == last_x:
                break

            last_x = self.cursor

            current_char = egc_chr(self._box.canvas[0, self.cursor]["ord"])
            if not first_char_found:
                if not current_char.isspace():
                    first_char_found = True
                    char_is_word_char = is_word_char(current_char)
            elif current_char.isspace() or char_is_word_char != is_word_char(
                current_char
            ):
                break

    def _enter(self):
        if self.enter_callback is not None:
            self.enter_callback(self)

    def _backspace(self):
        if self.has_nonempty_selection:
            self._move_undo_buffer_to_stack("del")
            self._undo_buffer.append(self.delete_selection())
        else:
            if self._undo_buffer_type != "del":
                self._move_undo_buffer_to_stack("del")

            end = self.cursor
            self.move_cursor_left()
            start = self.cursor
            self.cursor = end
            if start != end:
                self._undo_buffer.append(self._del_text(start, end))

    def _delete(self):
        if self.has_nonempty_selection:
            self._move_undo_buffer_to_stack("del")
            self._undo_buffer.append(self.delete_selection())
        else:
            if self._undo_buffer_type != "del":
                self._move_undo_buffer_to_stack("del")

            start = self.cursor
            self.move_cursor_right()
            end = self.cursor
            self.cursor = start
            if start != end:
                self._undo_buffer.append(self._del_text(start, end))

    def _left(self):
        if self.has_nonempty_selection:
            select_start = min(self._selection_start, self._selection_end)
            self.unselect()
            self.cursor = select_start
        else:
            self.unselect()
            self.move_cursor_left()

    def _right(self):
        if self.has_nonempty_selection:
            select_end = max(self._selection_start, self._selection_end)
            self.unselect()
            self.cursor = select_end
        else:
            self.unselect()
            self.move_cursor_right()

    def _ctrl_left(self):
        self.unselect()
        self.move_word_left()

    def _ctrl_right(self):
        self.unselect()
        self.move_word_right()

    def _ctrl_a(self):
        """Select all."""
        self._selection_start = 0
        self._selection_end = self._line_width
        self.cursor = self._line_width

    def _ctrl_d(self):
        """Select word."""
        self.unselect()
        last_x = self.cursor
        while True:
            self.move_cursor_left()
            if last_x == self.cursor:
                break
            if not is_word_char(egc_chr(self._box.canvas[0, self.cursor]["ord"])):
                self.move_cursor_right()
                break
            last_x = self.cursor

        self.select()
        last_x = self.cursor
        while True:
            if not is_word_char(egc_chr(self._box.canvas[0, self.cursor]["ord"])):
                break
            self.move_cursor_right()
            if last_x == self.cursor:
                break
            last_x = self.cursor

    def _home(self):
        self.unselect()
        self.cursor = 0

    def _end(self):
        self.unselect()
        self.cursor = self._line_width

    def _shift_left(self):
        self.select()
        self.move_cursor_left()

    def _shift_right(self):
        self.select()
        self.move_cursor_right()

    def _shift_ctrl_left(self):
        self.select()
        self.move_word_left()

    def _shift_ctrl_right(self):
        self.select()
        self.move_word_right()

    def _shift_home(self):
        self.select()
        self.cursor = 0

    def _shift_end(self):
        self.select()
        self.cursor = self._line_width

    def _escape(self):
        if self.has_nonempty_selection:
            self.unselect()
            self._highlight_selection()
        else:
            self.unselect()
            self.blur()

    def _ascii(self, key):
        if self.has_nonempty_selection:
            self._move_undo_buffer_to_stack("add")
            self._undo_buffer.append(self.delete_selection())

        if (
            self.max_chars is None
            or grapheme_len(canvas_as_text(self._box.canvas[0, : self._line_width]))
            < self.max_chars
        ):
            if self._undo_buffer_type != "add":
                self._move_undo_buffer_to_stack("add")
            self._undo_buffer.append(self._add_text(self.cursor, key))

    __HANDLERS = {
        ("enter", False, False, False): _enter,
        ("backspace", False, False, False): _backspace,
        ("delete", False, False, False): _delete,
        ("left", False, False, False): _left,
        ("right", False, False, False): _right,
        ("left", False, True, False): _ctrl_left,
        ("right", False, True, False): _ctrl_right,
        ("home", False, False, False): _home,
        ("end", False, False, False): _end,
        ("left", False, False, True): _shift_left,
        ("right", False, False, True): _shift_right,
        ("left", False, True, True): _shift_ctrl_left,
        ("right", False, True, True): _shift_ctrl_right,
        ("home", False, False, True): _shift_home,
        ("end", False, False, True): _shift_end,
        ("escape", False, False, False): _escape,
        ("z", False, True, False): undo,
        ("z", False, True, True): redo,
        ("y", False, True, False): redo,  # ctrl + shift + z won't work on linux
        ("r", False, True, False): redo,
        ("a", False, True, False): _ctrl_a,
        ("d", False, True, False): _ctrl_d,
    }

    def on_key(self, key_event: KeyEvent) -> bool | None:
        """Process key press."""
        if not self.is_focused:
            return super().on_key(key_event)

        if (
            not key_event.alt
            and not key_event.ctrl
            and not key_event.shift
            and len(key_event.key) == 1
        ):
            self._ascii(key_event.key)
        elif handler := self.__HANDLERS.get(astuple(key_event)):
            handler(self)
        else:
            return super().on_key(key_event)

        return True

    def on_paste(self, paste_event: PasteEvent) -> bool | None:
        """Add paste to textbox."""
        if not self.is_focused:
            return

        self._move_undo_buffer_to_stack()
        undos = []
        if undo := self.delete_selection():
            undos.append(undo)
        undos.append(self._add_text(self.cursor, paste_event.paste))
        self._undo_stack.append(undos)
        self._redo_stack.clear()

        return True

    def grab(self, mouse_event):
        """Start selection on grab."""
        if mouse_event.button == "left" and self._box.collides_point(mouse_event.pos):
            super().grab(mouse_event)

            _, x = self._box.to_local(mouse_event.pos)

            if not mouse_event.shift:
                self.unselect()

            self.cursor = min(x, self._line_width)
            self.select()  # Need at least an empty selection for `grab_update`.

    def grab_update(self, mouse_event: MouseEvent):
        """Update selection on grab update."""
        if self._box.collides_point(mouse_event.pos):
            _, x = self._box.to_local(mouse_event.pos)
            self.cursor = min(x, self._line_width)
        else:
            _, x = self.to_local(mouse_event.pos)

            if x < 0:
                self.move_cursor_left()
            elif x >= self.width:
                self.move_cursor_right()

    def ungrab(self, mouse_event):
        """Clear an empty selection on ungrab."""
        super().ungrab(mouse_event)
        if self._selection_start == self._selection_end:
            self.unselect()



r"""
Support for loading and rendering FIGfonts.

A `FIGFont` (class) represents a FIGfont (file) which describe how to render ascii art
from normal text. `FIGFont.render_array` will render the ascii art into a numpy array
that can be copied into a `Text` canvas. `FIGFont.render_str` will render the ascii art
into a multiline string.

References
----------
- http://www.figlet.org/
- http://www.jave.de/figlet/figfont.html
- https://github.com/cmatsuoka/figlet
- https://github.com/pwaller/pyfiglet

See Also
--------
- http://www.figlet.org/fontdb.cgi
- https://github.com/salt-die/fig-fonts
"""

import re
import zipfile
from dataclasses import dataclass, field, fields
from enum import IntFlag
from itertools import islice
from pathlib import Path
from typing import Self

import numpy as np
from numpy.typing import NDArray
from ugrapheme import grapheme_iter
from uwcwidth import wcswidth

from .text_tools import egc_ord

__all__ = ["FullLayout", "FIGFont"]


class FullLayout(IntFlag):
    r"""
    Character layout for a FIGFont.

    The layout modes are:

    * **FullWidth**: Each character occupies the full width or height of its arrangement
      of sub-characters.
    * **Kerning**: Each character is moved together until they touch.
    * **Smushing**: Each character is moved one step closer after they touch, so that
      they overlap. Additional smushing rules determine which sub-character is used for
      each overlap.

    There are two types of smushing:

    * **Universal**: The sub-character from the earlier character is replaced by the
      sub-character from the later character. (This behavior can be reversed with
      `reverse_universal_smush`)
    * **Controlled**: Uses a set of smushing rules.

    The controlled smushing rules are:

    * **Equal**: Two sub-characters are smushed into a single sub-character if they are
      equal (except for hardblanks).
    * **Underscore**: An underscore (``"_"``) will be replaced by any of: ``"|"``,
      ``"/"``, ``"\"``, ``"["``, ``"]"``, ``"{"``, ``"}"``, ``"("``, ``")"``, ``"<"``,
      ``">"``.
    * **Hierarchy**: A hierarchy of six classes is used: ``"|"``, ``"/\"``, ``"[]"``,
      ``"{}"``, ``"()"``, and ``"<>"``. When two sub-characters are from different
      classes, the latter class will be used.
    * **Pair**: Replaces opposite brackets (``"[]"`` or ``"]["``), braces (``"{}"`` or
      ``"}{"``), and parentheses (``"()"`` or ``")("``) with a vertical bar (``"|"``).
    * **BigX**: Replaces ``"/\"`` with ``"|"``, ``"\/"`` with ``"Y"``, and ``"><"``
      into ``"X"``.
    * **HardBlank**: Two hardblanks will be replaced with a single hardblank.
    """

    FullWidth = 0
    Equal = 1
    Underscore = 2
    Hierarchy = 4
    Pair = 8
    BigX = 16
    HardBlank = 32
    Kerning = 64
    Universal = 128
    # TODO: Add vertical smushing.

    @classmethod
    def from_old_layout(cls, old_layout: int) -> Self:
        """Return a full layout from an old layout."""
        return {-1: cls.FullWidth, 0: cls.Kerning}.get(
            old_layout, cls(old_layout | 128)
        )


@dataclass
class FIGFont:
    """
    An object representation of a FIGfont.

    Parameters
    ----------
    hardblank : str, default: "$"
        This character is used to represent whitespace that shouldn't be smushed.
    reverse_text : bool, default: False
        If true, text will be rendered right-to-left.
    layout : FullLayout, default: FullLayout.Universal
        Controls how characters are fitted in rendered text.
    reverse_universal_smush : bool, default: False
        Whether universal smushing will display earliest sub-character.
    font : dict[str, NDArray[np.dtype("<U1")]], default: {}
        A dictionary of characters to their ascii art representations.
    comments : str, default: ""
        Additional comments about this font.

    Attributes
    ----------
    hardblank : str
        This character is used to represent whitespace that shouldn't be smushed.
    reverse_text : bool
        If true, text will be rendered right-to-left.
    layout : FullLayout.Universal
        Controls how characters are fitted in rendered text.
    reverse_universal_smush : bool
        Whether universal smushing will display earliest sub-character.
    font : dict[str, NDArray[np.dtype("<U1")]]
        A dictionary of characters to their ascii art representations.
    comments : str
        Additional comments about this font.

    Methods
    -------
    from_path(path)
        Load a FIGFont from a path.
    render_array(text)
        Render text as ascii art into a 2-D "<U1" numpy array.
    render_str(text)
        Render text as ascii art into a multiline string.
    """

    hardblank: str = "$"
    """This character is used to represent whitespace that shouldn't be smushed."""
    reverse_text: bool = False
    """If true, text will be rendered right-to-left."""
    layout: FullLayout = FullLayout.Universal
    """Controls how characters are fitted in rendered text."""
    reverse_universal_smush: bool = False
    """Whether universal smushing will display earliest sub-character."""
    font: dict[str, NDArray[np.dtype("<U1")]] = field(repr=False, default_factory=dict)
    """A dictionary of characters to their ascii art representations."""
    comments: str = field(repr=False, default="")
    """Additional comments about this font."""

    @classmethod
    def from_dict(cls, attrs: dict) -> Self:
        """Return a FIGFont from a dictionary."""
        return cls(
            **{
                field.name: attrs[field.name]
                for field in fields(cls)
                if field.init and field.name in attrs
            }
        )

    @classmethod
    def from_path(cls, path: Path) -> Self:
        """Load a FIGFont from a path."""
        HEADER_RE = (
            r"^[tf]lf2.(?P<hardblank>.) (?P<height>\d+) \d+ \d+ "
            r"(?P<old_layout>-?\d+) (?P<comment_lines>\d+)"
            r"(?: (?P<reverse_text>0|1)"
            r"(?: (?P<layout>\d+)"
            r"(?: \d+)?)?)?\s*$"
        )
        ENDMARKS_RE = re.compile(r"(\S)\1*\s*$")
        NUMBER_RE = re.compile(r"^(0[0-7]*|0x[a-fA-F0-9]+|[1-9]\d*)(?:\s+\w*)?$")

        if zipfile.is_zipfile(path):
            with zipfile.ZipFile(path) as f:
                figdata = f.open(f.namelist()[0]).read()
        else:
            figdata = path.read_bytes()

        header, *lines = figdata.decode(errors="ignore").splitlines()

        if not (m := re.match(HEADER_RE, header)):
            raise ValueError("Invalid FIGfont file.")

        figinfo = m.groupdict()
        for k, v in figinfo.items():
            if isinstance(v, str) and v.removeprefix("-").isdigit():
                figinfo[k] = int(v)

        figinfo["reverse_text"] = bool(figinfo["reverse_text"])

        if figinfo["layout"] is None:
            figinfo["layout"] = FullLayout.from_old_layout(figinfo["old_layout"])
        else:
            figinfo["layout"] = FullLayout(figinfo["layout"])

        figinfo["comments"] = "\n".join(lines[: figinfo["comment_lines"]])
        del lines[: figinfo["comment_lines"]]

        it = iter(lines)
        height = figinfo["height"]

        def consume_char() -> NDArray[np.dtype("<u1")] | None:
            char_lines = [ENDMARKS_RE.sub("", line) for line in islice(it, height)]
            if len(char_lines) < height:
                return None

            width = max(wcswidth(line) for line in char_lines)
            if width < 0:
                return None

            char = np.full((height, width), " ")
            for i, line in enumerate(char_lines):
                j = 0
                for grapheme in grapheme_iter(line):
                    cwidth = wcswidth(grapheme)
                    if cwidth == 0:
                        continue

                    char.view(np.uint32)[i, j] = egc_ord(grapheme)
                    if cwidth > 1:
                        char[i, j + 1 : j + cwidth] = ""

                    j += cwidth

            return char

        font = {}
        for i in range(32, 127):
            font[chr(i)] = consume_char()

        for char in "ÄÖÜäöüß":
            font[char] = consume_char()

        # ! No support for control files, so remaining code tags will be interpreted as
        # ! unicode code points (negative code tags will be skipped).
        for line in it:
            if not (m := NUMBER_RE.match(line)):
                continue

            cap = m.group(1)
            if cap.startswith("0x"):
                code_tag = int(cap, 16)
            elif cap.startswith("0"):
                code_tag = int(cap, 8)
            else:
                code_tag = int(cap)

            if (fig_char := consume_char()) is not None:
                font[chr(code_tag)] = fig_char

        figinfo["font"] = font
        return cls.from_dict(figinfo)

    @property
    def height(self) -> int:
        """Height of characters in this font."""
        return next(v for v in self.font.values() if v is not None).shape[0]

    def _trim_char(
        self, fig_char: NDArray[np.dtype("<U1")]
    ) -> NDArray[np.dtype("<U1")]:
        """Remove leading and trailing whitespace."""
        while fig_char.shape[1] and (fig_char[:, 0] == " ").all():
            fig_char = fig_char[:, 1:]

        while fig_char.shape[1] and (fig_char[:, -1] == " ").all():
            fig_char = fig_char[:, :-1]

        return fig_char

    def _smush_subchar(self, a: str, b: str) -> str | None:
        """
        Attempt to smush two sub-characters given the current layout.
        If smushing fails, return None, else return the smushed sub-character.
        """
        if a.isspace():
            return b

        if b.isspace():
            return a

        # Universal smushing
        if not self.layout & 63:
            if a == self.hardblank:
                return b

            if b == self.hardblank:
                return a

            return a if self.reverse_text ^ self.reverse_universal_smush else b

        if (
            self.layout & FullLayout.HardBlank
            and a == self.hardblank
            and b == self.hardblank
        ):
            return a

        if a == self.hardblank or b == self.hardblank:
            return None

        if self.layout & FullLayout.Equal and a == b:
            return a

        smushes = []
        if self.layout & FullLayout.Underscore:
            smushes.append(("_", "|/\\[]{}()<>"))

        if self.layout & FullLayout.Hierarchy:
            smushes.extend(
                (
                    ("|", "|/\\[]{}()<>"),
                    ("\\/", "[]{}()<>"),
                    ("[]", "{}()<>"),
                    ("{}", "()<>"),
                    ("()", "<>"),
                )
            )

        for low, high in smushes:
            if a in low and b in high:
                return b
            if b in low and a in high:
                return a

        ab = b + a if self.reverse_text else a + b
        if self.layout & FullLayout.Pair and ab in {"[]", "][", "()", ")(", "{}", "}{"}:
            return "|"

        if self.layout & FullLayout.BigX:
            if ab == "/\\":
                return "|"
            if ab == "\\/":
                return "Y"
            if ab == "><":
                return "X"

    def _smush(
        self, a: NDArray[np.dtype("<U1")], b: NDArray[np.dtype("<U1")]
    ) -> list[str] | None:
        """
        Attempt to smush two columns of sub-characters.

        If smushing fails, return None, else return the smushed column as a list of
        characters.
        """
        c = []
        for sub_a, sub_b in zip(a, b):
            if sub_c := self._smush_subchar(sub_a, sub_b):
                c.append(sub_c)
            else:
                return None
        return c

    def _add_char(
        self, buffer: NDArray[np.dtype("<U1")], prev_char_width: int, char: str
    ) -> tuple[NDArray[np.dtype("<U1")], int]:
        """Add a character to the line buffer."""
        fig_char = self.font.get(char, self.font.get("\x00"))
        if fig_char is None:
            return buffer, 0
        else:
            fig_char = fig_char.copy()

        if self.layout:
            fig_char = self._trim_char(fig_char)
        current_char_width = fig_char.shape[1]
        a, b = (fig_char, buffer) if self.reverse_text else (buffer, fig_char)

        # If characters are wide enough and any smushing rule (controlled or universal)
        # is enabled, attempt to smush the last column of a with the first column of b.
        if (
            prev_char_width >= 2
            and current_char_width >= 2
            and self.layout & 191
            and (c := self._smush(a[:, -1], b[:, 0]))
        ):
            a[:, -1] = c
            b = b[:, 1:]

        return np.concatenate((a, b), axis=1), current_char_width

    def _render_line(self, line: str) -> NDArray[np.dtype("<U1")]:
        """Render a single line of text."""
        buffer, prev_char_width = np.zeros((self.height, 0), dtype=str), 0
        for char in line:
            buffer, prev_char_width = self._add_char(buffer, prev_char_width, char)

        buffer[buffer == self.hardblank] = " "
        return buffer

    def render_array(self, text: str) -> NDArray[np.dtype("<U1")]:
        """
        Render text as ascii art into a 2-D "<U1" numpy array.

        Parameters
        ----------
        text : str
            Text to render as ascii art into an array.

        Returns
        -------
        NDArray[np.dtype("<U1")]
            The rendered array.
        """
        lines = list(map(self._render_line, text.splitlines()))
        max_width = max(line.shape[1] for line in lines)
        for i, line in enumerate(lines):
            if line.shape[1] != max_width:
                lines[i] = np.pad(
                    line,
                    ((0, 0), (0, max_width - line.shape[1])),
                    constant_values=" ",
                )

        # TODO: Add vertical smushing.

        return np.concatenate(lines, axis=0)

    def render_str(self, text: str) -> str:
        """
        Render text as ascii art into a multiline string.

        Parameters
        ----------
        text : str
            Text to render as ascii art into a multiline string.

        Returns
        -------
        str
            The rendered string.
        """
        return "\n".join("".join(line) for line in self.render_array(text))



"""A slider gadget."""

from collections.abc import Callable

from ..colors import BLACK, WHITE, Color
from ..terminal.events import MouseButton, MouseEvent
from .behaviors.grabbable import Grabbable
from .pane import Pane, Point, PosHint, Size, SizeHint, bindable, clamp
from .text import Text

__all__ = ["Slider", "Point", "Size"]


class Slider(Grabbable, Pane):
    r"""
    A slider gadget.

    Parameters
    ----------
    min : float
        Minimum value of slider.
    max : float
        Maximum value of slider.
    start_value: float | None, default: None
        Start value of slider. If `None`, start value is :attr:`min`.
    callback : Callable | None, default: None
        Single argument callable called with new value of slider when slider is updated.
    handle_color : Color, default: WHITE
        Color of handle.
    slider_color: Color, default: WHITE
        Color of slider.
    fill_color: Color, default: WHITE
        Color of filled portion of slider.
    slider_enabled : bool, default: True
        Whether slider value can be changed.
    is_grabbable : bool, default: True
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool, default: False
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton, default: "left"
        Mouse button used for grabbing.
    bg_color : Color, default: BLACK
        Background color of gadget.
    alpha : float, default: 1.0
        Transparency of gadget.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    min : float
        Minimum value of slider.
    max : float
        Maximum value of slider.
    value : float
        Current value of slider.
    callback : Callable
        Single argument callable called with new value of slider when slider is updated.
    slider_color: Color
        Color of slider.
    fill_color: Color
        Color of filled portion of slider.
    fill_color: Color
        Color of filled portion of slider.
    slider_enabled : bool
        Whether slider value can be changed.
    proportion : float
        Current proportion of slider.
    is_grabbable : bool
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton
        Mouse button used for grabbing.
    bg_color : Color
        Background color of gadget.
    alpha : float
        Transparency of gadget.
    is_grabbed : bool
        Whether gadget is grabbed.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    grab(mouse_event)
        Grab the gadget.
    ungrab(mouse_event)
        Ungrab the gadget.
    grab_update(mouse_event)
        Update gadget with incoming mouse events while grabbed.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        min: float,
        max: float,
        start_value: float | None = None,
        callback: Callable | None = None,
        handle_color: Color = WHITE,
        slider_color: Color = WHITE,
        fill_color: Color = WHITE,
        slider_enabled: bool = True,
        is_grabbable: bool = True,
        ptf_on_grab: bool = False,
        mouse_button: MouseButton = "left",
        bg_color: Color = BLACK,
        alpha: float = 1.0,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            is_grabbable=is_grabbable,
            ptf_on_grab=ptf_on_grab,
            mouse_button=mouse_button,
            bg_color=bg_color,
            alpha=alpha,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        if min >= max:
            raise ValueError(f"{min=} >= {max=}")
        self._min = min
        self._max = max

        self._slider = Text(
            default_cell="━",
            size=(1, 1),
            pos_hint={"y_hint": 0.5, "anchor": "top"},
            size_hint={"width_hint": 1.0},
            is_transparent=True,
        )
        self._handle = Text(
            default_cell="█", size=(1, 1), pos_hint={"y_hint": 0.5, "anchor": "top"}
        )
        self.add_gadgets(self._slider, self._handle)

        self.handle_color = handle_color
        self.slider_color = slider_color
        self.fill_color = fill_color
        self.callback = callback
        self.slider_enabled = slider_enabled
        self.value = self.min if start_value is None else start_value

    @property
    def min(self) -> float:
        """Minimum value of slider."""
        return self._min

    @min.setter
    def min(self, value: float):
        if value >= self.max:
            raise ValueError("Min can't be greater than or equal to max.")
        self._min = value
        self.proportion = self.proportion

    @property
    def max(self) -> float:
        """Maximum value of slider."""
        return self._max

    @max.setter
    def max(self, value: float):
        if value <= self.min:
            raise ValueError("Max can't be less than or equal to min.")
        self._max = value
        self.proportion = self.proportion

    @property
    def handle_color(self) -> Color:
        """Color of handle."""
        return self._handle.default_fg_color

    @handle_color.setter
    def handle_color(self, handle_color: Color | str):
        self._handle.default_fg_color = handle_color
        self._handle.canvas["fg_color"] = handle_color

    @property
    def slider_color(self) -> Color:
        """Color of slider."""
        return self._slider.default_fg_color

    @slider_color.setter
    def slider_color(self, slider_color: Color | str):
        self._slider.default_fg_color = slider_color
        self._slider.canvas["fg_color"][:, self._handle.x :] = slider_color

    @property
    def fill_color(self) -> Color:
        """Color of "filled" portion of slider."""
        return self._fill_color

    @fill_color.setter
    def fill_color(self, color: Color):
        self._fill_color = color
        self._slider.canvas["fg_color"][:, : self._handle.x] = color

    @property
    def proportion(self) -> float:
        """Current proportion of slider."""
        return self._proportion

    @proportion.setter
    @bindable
    def proportion(self, value: float):
        if not self.slider_enabled:
            return

        self._proportion = clamp(value, 0, 1)
        self._value = (self.max - self.min) * self._proportion + self.min

        self._handle.x = x = round(self._proportion * self.fill_width)
        self._slider.canvas["fg_color"][:, :x] = self.fill_color
        self._slider.canvas["fg_color"][:, x:] = self.slider_color

        if self.callback is not None:
            self.callback(self._value)

    @property
    def value(self) -> float:
        """Current value of slider."""
        return self._value

    @value.setter
    @bindable
    def value(self, value: float):
        value = clamp(value, self.min, self.max)
        self.proportion = (value - self.min) / (self.max - self.min)

    @property
    def fill_width(self):
        """Width of the slider minus the width of the handle."""
        return self.width - self._handle.width

    def on_size(self):
        """Resize canvas and color arrays and reposition slider handle."""
        super().on_size()
        self.proportion = self.proportion

    def on_add(self):
        """Resize canvas and color arrays and reposition slider handle."""
        super().on_add()
        self.proportion = self.proportion

    def grab(self, mouse_event: MouseEvent):
        """Move handle to mouse position on grab."""
        if (
            mouse_event.event_type == "mouse_down"
            and self.collides_point(mouse_event.pos)
            and self.to_local(mouse_event.pos).y == self.height // 2
        ):
            super().grab(mouse_event)
            self.grab_update(mouse_event)

    def grab_update(self, mouse_event: MouseEvent):
        """Update proportion and handle position on grab update."""
        x = clamp(self.to_local(mouse_event.pos).x, 0, self.fill_width)
        self.proportion = 0 if self.fill_width == 0 else x / self.fill_width



"""Draggable horizontal and vertical split layouts."""

from time import perf_counter
from typing import Literal

from ..colors import WHITE, Color, lerp_colors
from ..geometry.easings import in_quart, out_quart
from .behaviors.grabbable import Grabbable
from .gadget import Gadget, Point, PosHint, Size, SizeHint, clamp
from .pane import Pane

__all__ = ["HSplitLayout", "VSplitLayout", "Point", "Size"]

GRAY = Color(127, 127, 127)


class _Handle(Grabbable, Pane):
    def __init__(self, size_hint):
        super().__init__(size=(2, 2), size_hint=size_hint, alpha=0)

    def on_mouse(self, mouse_event):
        self.alpha = (self.is_grabbed or self.collides_point(mouse_event.pos)) * 0.5
        return super().on_mouse(mouse_event)

    @property
    def bg_color(self) -> Color:
        p = perf_counter() % 1.0
        if p < 0.5:
            return lerp_colors(GRAY, WHITE, in_quart(2.0 * p))
        return lerp_colors(WHITE, GRAY, out_quart(2.0 * p - 1.0))

    @bg_color.setter
    def bg_color(self, _):
        pass


class _HSplitHandle(_Handle):
    def grab_update(self, mouse_event):
        self.parent: HSplitLayout
        self.parent.split_row += mouse_event.dy


class _VSplitHandle(_Handle):
    def grab_update(self, mouse_event):
        self.parent: VSplitLayout
        self.parent.split_col += mouse_event.dx


class HSplitLayout(Gadget):
    r"""
    A horizontal split layout.

    Add gadgets to the :attr:``top_pane`` or :attr:``bottom_pane``, e.g.,
    ``my_hsplit.top_pane.add_gadget(my_gadget)``.

    If :attr:``top_min_height`` and :attr:``bottom_min_height`` can't both be satisfied,
    :attr:``split_anchor`` will determine which pane's height requirement is preferred.

    Parameters
    ----------
    split_row : int, default: 1
        Height of top pane. Height of bottom pane is ``self.width - split_row``.
    split_anchor : Literal["top", "bottom"], default: "top
        If gadget is resized, ``split_anchor`` determines which pane's height isn't
        changed.
    split_resizable : bool, default: True
        Whether split is resizable with a grabbable handle.
    top_min_height : int, default: 1
        Minimum height of top pane.
    bottom_min_height : int, default: 1
        Minimum height of bottom pane.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    top_pane : Gadget
        Container gadget for top side of split layout.
    bottom_pane : Gadget
        Container gadget for bottom side of split layout.
    split_row : int
        Height of top pane. Height of bottom pane is ``self.width - split_row``.
    split_anchor : Literal["top", "bottom"]
        If gadget is resized, ``split_anchor`` determines which pane's height isn't
        changed.
    split_resizable : bool
        Whether split is resizable with a grabbable handle.
    top_min_height : int
        Minimum height of top pane.
    bottom_min_height : int
        Minimum height of bottom pane.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        split_row: int = 1,
        split_anchor: Literal["top", "bottom"] = "top",
        split_resizable: bool = True,
        top_min_height: int = 1,
        bottom_min_height: int = 1,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self.top_pane = Gadget(size_hint={"width_hint": 1.0})
        """Container gadget for top side of split layout."""
        self.bottom_pane = Gadget(size_hint={"width_hint": 1.0})
        """Container gadget for bottom side of split layout."""
        self.handle = _HSplitHandle({"width_hint": 1.0})
        """Grabbable handle gadget used to resize split layout."""

        def adjust():
            self.bottom_pane.top = self.top_pane.bottom
            self.handle.top = self.top_pane.bottom - 1

        self.top_pane.bind("size", adjust)
        self.add_gadgets(self.top_pane, self.bottom_pane, self.handle)

        self.split_anchor = split_anchor
        """
        If gadget is resized, ``split_anchor`` determines which pane's height isn't
        changed.
        """
        self.split_resizable = split_resizable
        """Whether split is resizable with a grabbable handle."""
        self._top_min_height = top_min_height
        """Minimum height of top pane."""
        self._bottom_min_height = bottom_min_height
        """Minimum height of bottom pane."""
        self.split_row = split_row
        """Height of top pane."""

    @property
    def top_min_height(self) -> int:
        """Minimum height of top pane."""
        return self._top_min_height

    @top_min_height.setter
    def top_min_height(self, top_min_height: int):
        self._top_min_height = clamp(top_min_height, 0, None)
        self.split_row = self.split_row

    @property
    def bottom_min_height(self) -> int:
        """Minimum height of bottom pane."""
        return self._bottom_min_height

    @bottom_min_height.setter
    def bottom_min_height(self, bottom_min_height: int):
        self._bottom_min_height = clamp(bottom_min_height, 0, None)
        self.split_row = self.split_row

    @property
    def split_row(self) -> int:
        """Height of top pane."""
        return self.top_pane.height

    @split_row.setter
    def split_row(self, split_row: int):
        if (
            self.top_min_height <= split_row
            and self.bottom_min_height <= self.height - split_row
        ):
            self.top_pane.height = split_row
            self.bottom_pane.height = self.height - split_row
        elif split_row < self.top_min_height:
            if (
                self.split_anchor == "top"
                or self.bottom_min_height <= self.height - self.top_min_height
            ):
                self.top_pane.height = self.top_min_height
                self.bottom_pane.height = self.height - self.top_min_height
            else:
                self.bottom_pane.height = self.bottom_min_height
                self.top_pane.height = self.height - self.bottom_min_height
        elif (
            self.split_anchor == "bottom"
            or self.top_min_height <= self.height - self.bottom_min_height
        ):
            self.bottom_pane.height = self.bottom_min_height
            self.top_pane.height = self.height - self.bottom_min_height
        else:
            self.top_pane.height = self.top_min_height
            self.bottom_pane.height = self.height - self.top_min_height

    @property
    def split_resizable(self) -> bool:
        """Whether split is resizable with a grabbable handle."""
        return self.handle.is_grabbable

    @split_resizable.setter
    def split_resizable(self, split_resizable: bool):
        self.handle.is_grabbable = split_resizable

    def on_size(self):
        """Resize panes on resize."""
        if self.split_anchor == "top":
            self.split_row = self.split_row
        else:
            self.split_row = self.height - self.bottom_pane.height


class VSplitLayout(Gadget):
    r"""
    A vertical split layout.

    Add gadgets to the :attr:``left_pane`` or :attr:``right_pane``, e.g.,
    ``my_vsplit.left_pane.add_gadget(my_gadget)``.

    If :attr:``left_min_width`` and :attr:``right_min_width`` can't both be satisfied,
    :attr:``split_anchor`` will determine which pane's width requirement is preferred.

    Parameters
    ----------
    split_col : int, default: 1
        Width of left pane. Height of right pane is ``self.width - split_col``.
    split_anchor : Literal["left", "right"], default: "left"
        If gadget is resized, ``split_anchor`` determines which pane's width isn't
        changed.
    split_resizable : bool, default: True
        Whether split is resizable with a grabbable handle.
    left_min_width : int, default: 1
        Minimum width of left pane.
    right_min_width : int, default: 1
        Minimum width of right pane.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    left_pane : Gadget
        Container gadget for left side of split layout.
    right_pane : Gadget
        Container gadget for right side of split layout.
    split_col : int
        Width of left pane. Height of right pane is ``self.width - split_col``.
    split_anchor : Literal["left", "right"]
        If gadget is resized, ``split_anchor`` determines which pane's width isn't
        changed.
    split_resizable : bool
        Whether split is resizable with a grabbable handle.
    left_min_width : int
        Minimum width of left pane.
    right_min_width : int
        Minimum width of right pane.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        split_col: int = 1,
        split_anchor: Literal["left", "right"] = "left",
        split_resizable: bool = True,
        left_min_width: int = 1,
        right_min_width: int = 1,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self.left_pane = Gadget(size_hint={"height_hint": 1.0})
        """Container gadget for left side of split layout."""
        self.right_pane = Gadget(size_hint={"height_hint": 1.0})
        """Container gadget for right side of split layout."""
        self.handle = _VSplitHandle({"height_hint": 1.0})
        """Grabbable handle gadget used to resize split layout."""

        def adjust():
            self.right_pane.left = self.left_pane.right
            self.handle.left = self.left_pane.right - 1

        self.left_pane.bind("size", adjust)
        self.add_gadgets(self.left_pane, self.right_pane, self.handle)

        self.split_anchor = split_anchor
        """
        If gadget is resized, ``split_anchor`` determines which pane's width isn't
        changed.
        """
        self.split_resizable = split_resizable
        """ Whether split is resizable with a grabbable handle."""
        self._left_min_width = left_min_width
        """Minimum width of left pane."""
        self._right_min_width = right_min_width
        """Minimum width of right pane."""
        self.split_col = split_col
        """Width of left pane."""

    @property
    def left_min_width(self) -> int:
        """Minimum width of left pane."""
        return self._left_min_width

    @left_min_width.setter
    def left_min_width(self, left_min_width: int):
        self._left_min_width = clamp(left_min_width, 0, None)
        self.split_col = self.split_col

    @property
    def right_min_width(self) -> int:
        """Minimum width of right pane."""
        return self._right_min_width

    @right_min_width.setter
    def right_min_width(self, right_min_width: int):
        self._right_min_width = clamp(right_min_width, 0, None)
        self.split_col = self.split_col

    @property
    def split_col(self) -> int:
        """Width of left pane."""
        return self.left_pane.width

    @split_col.setter
    def split_col(self, split_col: int):
        if (
            self.left_min_width <= split_col
            and self.right_min_width <= self.width - split_col
        ):
            self.left_pane.width = split_col
            self.right_pane.width = self.width - split_col
        elif split_col < self.left_min_width:
            if (
                self.split_anchor == "left"
                or self.right_min_width <= self.width - self.left_min_width
            ):
                self.left_pane.width = self.left_min_width
                self.right_pane.width = self.width - self.left_min_width
            else:
                self.right_pane.width = self.right_min_width
                self.left_pane.width = self.width - self.right_min_width
        elif (
            self.split_anchor == "right"
            or self.left_min_width <= self.width - self.right_min_width
        ):
            self.right_pane.width = self.right_min_width
            self.left_pane.width = self.width - self.right_min_width
        else:
            self.left_pane.width = self.left_min_width
            self.right_pane.width = self.width - self.left_min_width

    @property
    def split_resizable(self) -> bool:
        """Whether split is resizable with a grabbable handle."""
        return self.handle.is_grabbable

    @split_resizable.setter
    def split_resizable(self, split_resizable: bool):
        self.handle.is_grabbable = split_resizable

    def on_size(self):
        """Resize panes on resize."""
        if self.split_anchor == "left":
            self.split_col = self.split_col
        else:
            self.split_col = self.width - self.right_pane.width



"""A black hole effect."""

import asyncio
from random import choice

import numpy as np
from numpy.typing import NDArray

from ...colors import BLACK, WHITE, Color, gradient, lerp_colors
from ...geometry import BezierCurve, Point, clamp, move_along_path, points_on_circle
from ..text import Text
from ..text_field import TextParticleField
from ._particle import Particle

STARS = "*✸✺✹✷✵✶⋆'.⬫⬪⬩⬨⬧⬦⬥"
UNSTABLE = "◦◎◉●◉◎◦"
UNSTABLE_COLORS = [
    Color.from_hex("ffcc0d"),
    Color.from_hex("ff7326"),
    Color.from_hex("ff194d"),
    Color.from_hex("bf2669"),
    Color.from_hex("702a8c"),
    Color.from_hex("049dbf"),
]
STAR_COLOR = Color.from_hex("4a4a4d")
STAR_GRADIENT = gradient(STAR_COLOR, WHITE, n=6)
TOP_COLOR = Color.from_hex("8a008a")
MIDDLE_COLOR = Color.from_hex("00d1ff")
RNG = np.random.default_rng()


async def black_hole_effect(text: Text):
    """
    Create a black hole effect.

    Parameters
    ----------
    text : Text
        The text gadget which to run the effect.

    Warnings
    --------
    Modifying `text` size while effect is running will break the effect.
    """
    field = TextParticleField(size_hint={"height_hint": 1.0, "width_hint": 1.0})
    field.particles_from_cells(text.canvas)

    all_particles = list(Particle.iter_from_field(field))

    positions = RNG.random((field.nparticles, 2)) * text.size
    for particle, position in zip(all_particles, positions):
        particle.final_ord = particle.cell["ord"]
        particle.final_fg_color = Color(*particle.cell["fg_color"].tolist())
        particle.final_pos = particle.pos
        particle.cell["ord"] = ord(choice(STARS))
        particle.cell["fg_color"] = choice(STAR_GRADIENT)
        particle.pos = position

    black_hole = TextParticleField(
        size_hint={"height_hint": 1.0, "width_hint": 1.0}, is_transparent=True
    )
    black_hole_radius = clamp(text.height // 3, 3, text.width // 3)
    black_hole_center = Point(text.height // 2, text.width // 2)
    nparticles = black_hole_radius * 3
    black_hole.particle_coords = field.particle_coords[-nparticles:]
    black_hole.particle_cells = field.particle_cells[-nparticles:]
    black_hole_particles = list(Particle.iter_from_field(black_hole))

    field.particle_coords = field.particle_coords[:-nparticles]
    field.particle_cells = field.particle_cells[:-nparticles]
    other_particles = all_particles[:-nparticles]

    circle_positions = points_on_circle(black_hole.nparticles, black_hole_radius)
    black_hole_positions = circle_positions.copy()
    black_hole_positions[:, 1] *= 2
    black_hole_positions += black_hole_center

    text.add_gadgets(field, black_hole)

    await _forming(black_hole_particles, black_hole_positions)

    rot = _rotating(black_hole, circle_positions, black_hole_center)
    rotate_task = asyncio.create_task(rot)

    await _consuming(other_particles, black_hole_positions, black_hole_center)

    rotate_task.cancel()

    await _collapsing(
        black_hole_particles,
        black_hole.particle_coords,
        black_hole_center,
        black_hole_radius,
    )

    field.particle_coords = np.vstack(
        [field.particle_coords, black_hole.particle_coords]
    )
    field.particle_cells = np.append(field.particle_cells, black_hole.particle_cells)

    await _point_char(black_hole, black_hole_center)
    text.remove_gadget(black_hole)

    await _exploding(all_particles, black_hole_center)

    text.remove_gadget(field)


async def _forming(particles: list[Particle], positions: NDArray[np.float32]):
    for particle in particles:
        particle.cell["ord"] = ord("✸")
        particle.cell["fg_color"] = WHITE

    paths = [
        [BezierCurve(np.array([particle.pos, position], float))]
        for particle, position in zip(particles, positions)
    ]

    speed = len(particles)
    motions = [
        move_along_path(particle, path, speed=speed, easing="in_out_sine")
        for particle, path in zip(particles, paths)
    ]

    tasks = []
    for motion in motions:
        tasks.append(asyncio.create_task(motion))
        await asyncio.sleep(1 / speed)

    await asyncio.gather(*tasks)


async def _rotating(
    black_hole: TextParticleField, positions: NDArray[np.float64], center: Point
):
    angles = np.linspace(0, 2 * np.pi, 100, endpoint=False)
    cos = np.cos(angles)
    sin = np.sin(angles)
    rot = np.array([[cos, -sin], [sin, cos]]).T
    i = 0

    while True:
        new_positions = positions @ rot[i]
        new_positions[:, 1] *= 2
        new_positions += center
        black_hole.particle_coords = new_positions
        i += 1
        i %= 100
        await asyncio.sleep(0.01)


async def _consuming(
    particles: list[Particle], positions: NDArray[np.float32], center: Point
):
    particles.sort(key=lambda p: (p.pos.y - center.y) ** 2 + (p.pos.x - center.x) ** 2)

    paths = [
        [BezierCurve(np.array([particle.pos, choice(positions), center]))]
        for particle in particles
    ]

    def create_fade(particle):
        start_color = particle.cell["fg_color"]

        def fade_particle(p):
            particle.cell["fg_color"] = lerp_colors(start_color, BLACK, p)

        return fade_particle

    motions = [
        move_along_path(
            particle, path, speed=20, easing="in_exp", on_progress=create_fade(particle)
        )
        for particle, path in zip(particles, paths)
    ]

    tasks = []
    nconsumes = clamp(int(len(particles) / 10), 2, 15)
    while motions:
        for motion in motions[:nconsumes]:
            tasks.append(asyncio.create_task(motion))
        motions = motions[nconsumes:]
        nconsumes += 1
        await asyncio.sleep(0.10)

    await asyncio.gather(*tasks)


async def _collapsing(
    particles: list[Particle],
    positions: NDArray[np.float32],
    center: Point,
    radius: int,
):
    new_pos = (positions - center).astype(float)
    new_pos *= (radius + 3) / radius
    new_pos += center

    paths = []
    for particle, position in zip(particles, new_pos):
        particle_path = [
            BezierCurve(np.array([particle.pos, position])),
            BezierCurve(np.array([position, center])),
        ]
        paths.append(particle_path)

    motions = [
        move_along_path(particle, path, speed=20, easing="in_exp")
        for particle, path in zip(particles, paths)
    ]

    await asyncio.gather(*motions)


async def _point_char(black_hole: TextParticleField, center: Point):
    black_hole.particle_coords = np.array([center]).astype(np.float64)
    black_hole.particle_cells = black_hole.particle_cells[:1]
    for _ in range(3):
        for char in UNSTABLE:
            black_hole.particle_cells[0]["ord"] = ord(char)
            black_hole.particle_cells[0]["fg_color"] = choice(UNSTABLE_COLORS)
            await asyncio.sleep(0.05)


async def _exploding(particles: list[Particle], center: Point):
    nearby_paths = []
    final_paths = []
    for particle in particles:
        particle.pos = center
        particle.cell["ord"] = particle.final_ord
        particle.cell["fg_color"] = choice(UNSTABLE_COLORS)
        near_point = choice(points_on_circle(6, 5))
        near_point[1] *= 2
        near_point += particle.final_pos
        nearby_paths.append([BezierCurve(np.array([particle.pos, near_point]))])
        final_paths.append([BezierCurve(np.array([near_point, particle.final_pos]))])

    nearby_motions = [
        move_along_path(particle, path, speed=20, easing="out_exp")
        for particle, path in zip(particles, nearby_paths)
    ]

    await asyncio.gather(*nearby_motions)

    def create_fade(particle):
        start_color = particle.cell["fg_color"]

        def fade_particle(p):
            particle.cell["fg_color"] = lerp_colors(
                start_color, particle.final_fg_color, p
            )

        return fade_particle

    final_motions = [
        move_along_path(
            particle,
            path,
            speed=7,
            easing="in_cubic",
            on_progress=create_fade(particle),
        )
        for particle, path in zip(particles, final_paths)
    ]

    await asyncio.gather(*final_motions)
